@@ -8,11 +8,12 @@ import (
 	"fmt"
 	"io"
 	"math/big"
-	"math/rand"
+	"math/rand/v2"
 	"os"
 	"path/filepath"
 	"strconv"
 	"strings"
+	"sync"
 	"time"
 )
 
@@ -22,6 +23,7 @@ var DebugFlags = [...]string{
 	"panic",
 	"parse",
 	"tokens",
+	"trace",
 	"types",
 }
 
@@ -38,10 +40,12 @@ type Config struct {
 	formatFloat bool // Whether format is floating-point.
 	origin      int
 	bigOrigin   *big.Int
-	seed        int64
-	debug       [len(DebugFlags)]bool
-	source      rand.Source
+	seed        uint64
+	debug       [len(DebugFlags)]int
+	traceLevel  int
+	source      *rand.ChaCha8
 	random      *rand.Rand
+	randLock    sync.Mutex
 	maxBits     uint           // Maximum length of an integer; 0 means no limit.
 	maxDigits   uint           // Above this size, ints print in floating format.
 	maxStack    uint           // Maximum call stack depth.
@@ -62,9 +66,9 @@ func (c *Config) init() {
 		c.output = os.Stdout
 		c.errOutput = os.Stderr
 		c.origin = 1
-		c.seed = time.Now().UnixNano()
+		c.seed = uint64(time.Now().UnixNano())
 		c.bigOrigin = big.NewInt(1)
-		c.source = rand.NewSource(c.seed)
+		c.source = rand.NewChaCha8(makeSeed(c.seed))
 		c.random = rand.New(c.source)
 		c.maxBits = 1e6
 		c.maxDigits = 1e4
@@ -108,7 +112,7 @@ func (c *Config) Format() string {
 	return c.format
 }
 
-// Format returns the formatting string for rationals.
+// RatFormat returns the formatting string for rationals.
 func (c *Config) RatFormat() string {
 	return c.ratFormat
 }
@@ -151,22 +155,38 @@ func (c *Config) FloatFormat() (verb byte, prec int, ok bool) {
 	return c.formatVerb, c.formatPrec, c.formatFloat
 }
 
-// Debug returns the value of the specified boolean debugging flag.
-func (c *Config) Debug(flag string) bool {
+// Debug returns the value of the specified debugging flag, -1 if the
+// flag is unknown.
+func (c *Config) Debug(flag string) int {
 	for i, f := range DebugFlags {
 		if f == flag {
 			return c.debug[i]
 		}
 	}
-	return false
+	return -1
+}
+
+// ClearDebug turns off all debugging flags.
+func (c *Config) ClearDebug() {
+	for i := range DebugFlags {
+		c.debug[i] = 0
+	}
+}
+
+// Tracing reports whether the tracing level is set at or above level.
+func (c *Config) Tracing(level int) bool {
+	return c.traceLevel >= level
 }
 
 // SetDebug sets the value of the specified boolean debugging flag.
 // It returns false if the flag is unknown.
-func (c *Config) SetDebug(flag string, state bool) bool {
+func (c *Config) SetDebug(flag string, state int) bool {
 	c.init()
 	for i, f := range DebugFlags {
 		if f == flag {
+			if flag == "trace" {
+				c.traceLevel = state
+			}
 			c.debug[i] = state
 			return true
 		}
@@ -208,16 +228,53 @@ func (c *Config) Random() *rand.Rand {
 	return c.random
 }
 
+// Source returns the underlying source for random numbers.
+func (c *Config) Source() *rand.ChaCha8 {
+	c.init()
+	return c.source
+}
+
 // RandomSeed returns the seed used to initialize the random number generator.
-func (c *Config) RandomSeed() int64 {
-	return c.seed
+func (c *Config) RandomSeed() uint64 {
+	c.LockRandom()
+	seed := c.seed
+	c.UnlockRandom()
+	return seed
+}
+
+// makeSeed copies seed a bunch of times to make a 32-byte seed for ChaCha8.
+func makeSeed(seed uint64) [32]byte {
+	// Seed is 32 bytes. Best we can do I guess.
+	word := [8]uint8{}
+	for i := range 64 / 8 {
+		word[i] = uint8(seed)
+		seed >>= 8
+	}
+	buf := [32]byte{}
+	for i := 0; i < 32; i += 8 {
+		copy(buf[i:], word[:])
+	}
+	return buf
 }
 
 // SetRandomSeed sets the seed for the random number generator.
-func (c *Config) SetRandomSeed(seed int64) {
+func (c *Config) SetRandomSeed(seed uint64) {
 	c.init()
+	c.LockRandom()
 	c.seed = seed
-	c.source.Seed(seed)
+	c.source.Seed(makeSeed(seed)) // All we can do for now.
+	c.UnlockRandom()
+}
+
+// LockRandom serializes access to the random number generator.
+// Needed because pfor can cause concurrent calls to the generator.
+func (c *Config) LockRandom() {
+	c.randLock.Lock()
+}
+
+// UnlockRandom releases the lock on the random number generator.
+func (c *Config) UnlockRandom() {
+	c.randLock.Unlock()
 }
 
 // MaxBits returns the maximum integer size to store, in bits.
@@ -226,7 +283,7 @@ func (c *Config) MaxBits() uint {
 	return c.maxBits
 }
 
-// MaxBits sets the maximum integer size to store, in bits.
+// SetMaxBits sets the maximum integer size to store, in bits.
 func (c *Config) SetMaxBits(digits uint) {
 	c.init()
 	c.maxBits = digits
