@@ -6,6 +6,7 @@ package exec
 
 import (
 	"fmt"
+	"strings"
 
 	"robpike.io/ivy/value"
 )
@@ -14,26 +15,58 @@ import (
 type Function struct {
 	IsBinary bool
 	Name     string
-	Left     string
-	Right    string
+	Left     value.Expr
+	Right    value.Expr
 	Body     []value.Expr
 	Locals   []string
 	Globals  []string
+	Source   string
+	// At time of definition; needed to parse saved source correctly.
+	Ibase int
+}
+
+// argProgString builds a string representation of arg, to be used in printing the
+// source to an op. If the argument is a vector, it needs special handling to get
+// parentheses and nesting.
+func argProgString(b *strings.Builder, arg value.Expr) {
+	switch expr := arg.(type) {
+	case *value.VarExpr:
+		b.WriteString(expr.ProgString())
+		return
+	case value.VectorExpr:
+		b.WriteRune('(')
+		for i, elem := range expr {
+			if i > 0 {
+				b.WriteRune(' ')
+			}
+			argProgString(b, elem)
+		}
+		b.WriteRune(')')
+	default:
+		b.WriteString(fmt.Sprintf("<unknown type in op print: %T>", arg))
+	}
 }
 
 func (fn *Function) String() string {
-	left := ""
+	var b strings.Builder
+	b.WriteString("op ")
 	if fn.IsBinary {
-		left = fn.Left + " "
+		argProgString(&b, fn.Left)
+		b.WriteRune(' ')
 	}
-	s := fmt.Sprintf("op %s%s %s =", left, fn.Name, fn.Right)
+	b.WriteString(fn.Name)
+	b.WriteRune(' ')
+	argProgString(&b, fn.Right)
+	b.WriteString(" = ")
 	if len(fn.Body) == 1 {
-		return s + " " + fn.Body[0].ProgString()
+		b.WriteString(fn.Body[0].ProgString())
+	} else {
+		for _, stmt := range fn.Body {
+			b.WriteString("\n\t")
+			b.WriteString(stmt.ProgString())
+		}
 	}
-	for _, stmt := range fn.Body {
-		s += "\n\t" + stmt.ProgString()
-	}
-	return s
+	return b.String()
 }
 
 func (fn *Function) EvalUnary(context value.Context, right value.Value) value.Value {
@@ -47,7 +80,7 @@ func (fn *Function) EvalUnary(context value.Context, right value.Value) value.Va
 	}
 	c.push(fn)
 	defer c.pop()
-	c.AssignLocal(1, right)
+	value.Assign(context, fn.Right, right, right)
 	v := value.EvalFunctionBody(c, fn.Name, fn.Body)
 	if v == nil {
 		value.Errorf("no value returned by %q", fn.Name)
@@ -66,8 +99,8 @@ func (fn *Function) EvalBinary(context value.Context, left, right value.Value) v
 	}
 	c.push(fn)
 	defer c.pop()
-	c.AssignLocal(1, left)
-	c.AssignLocal(2, right)
+	value.Assign(context, fn.Left, left, left)
+	value.Assign(context, fn.Right, right, right)
 	v := value.EvalFunctionBody(c, fn.Name, fn.Body)
 	if v == nil {
 		value.Errorf("no value returned by %q", fn.Name)
