@@ -69,7 +69,7 @@ func TestAll(t *testing.T) {
 			}
 			if !runTest(t, path, lineNum, shouldFail, input, output) {
 				errCount++
-				if errCount > 3 {
+				if errCount > 100 {
 					t.Fatal("too many errors")
 				}
 			}
@@ -177,6 +177,8 @@ func getText(t *testing.T, fileName string, lineNum int, shouldFail bool, lines
 }
 
 func reset() {
+	testConf = config.Config{}
+	testConf.SetFloatPrec(256)
 	testConf.SetFormat("")
 	testConf.SetMaxBits(1e9)
 	testConf.SetMaxDigits(1e4)
