@@ -6,6 +6,7 @@ package parse
 
 import (
 	"fmt"
+	"strings"
 
 	"robpike.io/ivy/exec"
 	"robpike.io/ivy/scan"
@@ -17,62 +18,66 @@ import (
 //	"op" name arg <eol>
 //	"op" name arg '=' statements <eol>
 //	"op" arg name arg '=' statements <eol>
-//	"opdelete" name arg <eol>
-//	"opdelete" arg name arg <eol>
 //
 // statements:
 //
 //	expressionList
 //	'\n' (expressionList '\n')+ '\n' # For multiline definition, ending with blank line.
-func (p *Parser) functionDefn() {
-	undefine := false
-	switch tok := p.next(); tok.Type {
-	case scan.Op:
-	case scan.OpDelete:
-		undefine = true
-	default:
-		p.errorf("unexpected %s", tok) // Cannot happen but be safe.
-	}
+func (p *Parser) functionDefn(start int) {
+	tok := p.need(scan.Op)
 	fn := new(exec.Function)
 	// Two identifiers means: op arg.
 	// Three identifiers means: arg op arg.
-	idents := make([]string, 2, 3)
-	idents[0] = p.need(scan.Identifier).Text
-	idents[1] = p.need(scan.Identifier).Text
-	if p.peek().Type == scan.Identifier {
-		idents = append(idents, p.next().Text)
+	// arg can be name or parenthesized list of args.
+	// We scan the op as an arg too, because we're not sure which one it is.
+	args := make([]value.Expr, 2, 3)
+	args[0] = p.funcArg()
+	args[1] = p.funcArg()
+	nameArg := args[0]
+	if p.peek().Type == scan.Identifier || p.peek().Type == scan.LeftParen {
+		nameArg = args[1]
+		args = append(args, p.funcArg())
 	}
-	// Undefine if so requested.
-	if undefine {
-		p.need(scan.EOF)
-		p.context.Undefine(idents[len(idents)-2], len(idents) == 3)
-		return
+	if x, ok := nameArg.(*value.VarExpr); ok {
+		fn.Name = x.Name
+	} else {
+		p.errorf("invalid function name: %v", nameArg.ProgString())
 	}
+
+	// Prepare to declare arguments.
+	argNames := make(map[string]bool)
+	declare := func(x *value.VarExpr) {
+		if x.Name == fn.Name {
+			p.errorf("argument name %q is function name", fn.Name)
+		}
+		if argNames[x.Name] {
+			p.errorf("multiple arguments named %q", x.Name)
+		}
+		argNames[x.Name] = true
+		p.context.Declare(x.Name)
+	}
+
 	// Install the function in the symbol table so recursive ops work. (As if.)
 	var installMap map[string]*exec.Function
-	if len(idents) == 3 {
-		if idents[1] == "o" { // Poor choice due to outer product syntax.
+	if len(args) == 3 {
+		if fn.Name == "o" { // Poor choice due to outer product syntax.
 			p.errorf(`"o" is not a valid name for a binary operator`)
 		}
 		fn.IsBinary = true
-		fn.Left = idents[0]
-		fn.Name = idents[1]
-		fn.Right = idents[2]
-		p.context.Declare(fn.Left)
-		p.context.Declare(fn.Right)
+		fn.Left = args[0]
+		fn.Right = args[2]
+		walkVars(fn.Left, declare)
+		walkVars(fn.Right, declare)
 		installMap = p.context.BinaryFn
 	} else {
-		fn.Name = idents[0]
-		fn.Right = idents[1]
-		p.context.Declare(fn.Right)
+		fn.Right = args[1]
+		walkVars(fn.Right, declare)
 		installMap = p.context.UnaryFn
 	}
-	if fn.Name == fn.Left || fn.Name == fn.Right {
-		p.errorf("argument name %q is function name", fn.Name)
-	}
+
 	// Define it, but prepare to undefine if there's trouble.
 	prevDefn := installMap[fn.Name]
-	p.context.Define(fn)
+	p.context.Define(fn) // Source will come at the end.
 	defer p.context.ForgetAll()
 	succeeded := false
 	defer func() {
@@ -85,7 +90,7 @@ func (p *Parser) functionDefn() {
 		}
 	}()
 
-	tok := p.next()
+	tok = p.next()
 	switch tok.Type {
 	case scan.Assign:
 		// Either one line:
@@ -127,14 +132,43 @@ func (p *Parser) functionDefn() {
 	default:
 		p.errorf("expected newline after function declaration, found %s", tok)
 	}
-	p.context.Define(fn)
+	// Was there a leading comment? If so, bind it to the saved textual definition
+	history := p.scanner.History()
+	for start > 0 {
+		if !strings.HasPrefix(strings.TrimSpace(history[start-1]), "#") {
+			break
+		}
+		start--
+	}
+	fn.Source = p.source(start, len(history))
+	// Remember the base so we can parse the source text again after a save.
+	fn.Ibase, _ = p.context.Config().Base()
 	funcVars(fn)
 	succeeded = true
-	if p.context.Config().Debug("parse") {
-		p.Printf("op %s %s %s = %s\n", fn.Left, fn.Name, fn.Right, tree(fn.Body))
+	p.context.Define(fn)
+	if p.context.Config().Debug("parse") > 0 {
+		p.Printf("op %s %s %s = %s\n", fn.Left.ProgString(), fn.Name, fn.Right.ProgString(), tree(fn.Body))
 	}
 }
 
+// function argument
+//	name | '(' args ')'
+func (p *Parser) funcArg() value.Expr {
+	tok := p.next()
+	if tok.Type == scan.Identifier {
+		return value.NewVarExpr(tok.Text)
+	}
+	if tok.Type != scan.LeftParen {
+		p.errorf("invalid function argument syntax at %s", tok.Text)
+	}
+	var v value.VectorExpr
+	for p.peek().Type != scan.RightParen {
+		v = append(v, p.funcArg())
+	}
+	p.next()
+	return v
+}
+
 // references returns a list, in appearance order, of the user-defined ops
 // referenced by this function body. Only the first appearance creates an
 // entry in the list.
@@ -143,13 +177,13 @@ func references(c *exec.Context, body []value.Expr) []exec.OpDef {
 	for _, expr := range body {
 		walk(expr, false, func(expr value.Expr, _ bool) {
 			switch e := expr.(type) {
-			case *unary:
-				if c.UnaryFn[e.op] != nil {
-					addReference(&refs, e.op, false)
+			case *value.UnaryExpr:
+				if c.UnaryFn[e.Op] != nil {
+					addReference(&refs, e.Op, false)
 				}
-			case *binary:
-				if c.BinaryFn[e.op] != nil {
-					addReference(&refs, e.op, true)
+			case *value.BinaryExpr:
+				if c.BinaryFn[e.Op] != nil {
+					addReference(&refs, e.Op, true)
 				}
 			}
 		})
@@ -183,31 +217,31 @@ func addReference(refs *[]exec.OpDef, name string, isBinary bool) {
 //	x = 1
 func funcVars(fn *exec.Function) {
 	known := make(map[string]int)
-	addLocal := func(name string) {
-		fn.Locals = append(fn.Locals, name)
-		known[name] = len(fn.Locals)
-	}
-	if fn.Left != "" {
-		addLocal(fn.Left)
-	}
-	if fn.Right != "" {
-		addLocal(fn.Right)
+	addLocal := func(e *value.VarExpr) {
+		fn.Locals = append(fn.Locals, e.Name)
+		known[e.Name] = len(fn.Locals)
 	}
 	f := func(expr value.Expr, assign bool) {
 		switch e := expr.(type) {
-		case *variableExpr:
-			x, ok := known[e.name]
+		case *value.VarExpr:
+			x, ok := known[e.Name]
 			if !ok {
 				if assign {
-					addLocal(e.name)
+					addLocal(e)
 				} else {
-					known[e.name] = 0
+					known[e.Name] = 0
 				}
-				x = known[e.name]
+				x = known[e.Name]
 			}
-			e.local = x
+			e.Local = x
 		}
 	}
+	if fn.Left != nil {
+		walk(fn.Left, true, f)
+	}
+	if fn.Right != nil {
+		walk(fn.Right, true, f)
+	}
 	for _, e := range fn.Body {
 		walk(e, false, f)
 	}
@@ -220,25 +254,25 @@ func funcVars(fn *exec.Function) {
 // after which it calls f(expr, assign).
 func walk(expr value.Expr, assign bool, f func(value.Expr, bool)) {
 	switch e := expr.(type) {
-	case *unary:
-		walk(e.right, false, f)
-	case conditional:
-		walk(e.binary, false, f)
-	case *binary:
-		walk(e.right, false, f)
-		walk(e.left, e.op == "=", f)
-	case *index:
-		for i := len(e.right) - 1; i >= 0; i-- {
-			x := e.right[i]
+	case *value.UnaryExpr:
+		walk(e.Right, false, f)
+	case *value.CondExpr:
+		walk(e.Cond, false, f)
+	case *value.BinaryExpr:
+		walk(e.Right, false, f)
+		walk(e.Left, e.Op == "=", f)
+	case *value.IndexExpr:
+		for i := len(e.Right) - 1; i >= 0; i-- {
+			x := e.Right[i]
 			if x != nil { // Not a placeholder index.
-				walk(e.right[i], false, f)
+				walk(e.Right[i], false, f)
 			}
 		}
-		walk(e.left, false, f)
-	case *variableExpr:
-	case sliceExpr:
+		walk(e.Left, false, f)
+	case *value.VarExpr:
+	case value.VectorExpr:
 		for i := len(e) - 1; i >= 0; i-- {
-			walk(e[i], false, f)
+			walk(e[i], assign, f)
 		}
 	case value.Char:
 	case value.Int:
@@ -246,10 +280,23 @@ func walk(expr value.Expr, assign bool, f func(value.Expr, bool)) {
 	case value.BigRat:
 	case value.BigFloat:
 	case value.Complex:
-	case value.Vector:
+	case *value.Vector:
 	case *value.Matrix:
 	default:
-		fmt.Printf("unknown %T in references\n", e)
+		fmt.Printf("unknown %T in walk\n", e)
 	}
 	f(expr, assign)
 }
+
+func walkVars(expr value.Expr, f func(*value.VarExpr)) {
+	switch e := expr.(type) {
+	case *value.VarExpr:
+		f(e)
+	case value.VectorExpr:
+		for i := len(e) - 1; i >= 0; i-- {
+			walkVars(e[i], f)
+		}
+	default:
+		fmt.Printf("unknown %T in variable list\n", e)
+	}
+}
