@@ -5,9 +5,8 @@
 package parse // import "robpike.io/ivy/parse"
 
 import (
-	"bytes"
 	"fmt"
-	"strconv"
+	"slices"
 	"strings"
 
 	"robpike.io/ivy/exec"
@@ -28,7 +27,7 @@ func tree(e interface{}) string {
 		return fmt.Sprintf("<float %s>", e)
 	case value.Complex:
 		return fmt.Sprintf("<complex %s>", e)
-	case sliceExpr:
+	case value.VectorExpr:
 		s := "<"
 		for i, x := range e {
 			if i > 0 {
@@ -38,17 +37,17 @@ func tree(e interface{}) string {
 		}
 		s += ">"
 		return s
-	case *variableExpr:
-		return fmt.Sprintf("<var %s>", e.name)
-	case *unary:
-		return fmt.Sprintf("(%s %s)", e.op, tree(e.right))
-	case *binary:
-		return fmt.Sprintf("(%s %s %s)", tree(e.left), e.op, tree(e.right))
-	case conditional:
-		return tree(e.binary)
-	case *index:
-		s := fmt.Sprintf("(%s[", tree(e.left))
-		for i, v := range e.right {
+	case *value.VarExpr:
+		return fmt.Sprintf("<var %s>", e.Name)
+	case *value.UnaryExpr:
+		return fmt.Sprintf("(%s %s)", e.Op, tree(e.Right))
+	case *value.BinaryExpr:
+		return fmt.Sprintf("(%s %s %s)", tree(e.Left), e.Op, tree(e.Right))
+	case *value.CondExpr:
+		return tree(e.Cond)
+	case *value.IndexExpr:
+		s := fmt.Sprintf("(%s[", tree(e.Left))
+		for i, v := range e.Right {
 			if i > 0 {
 				s += "; "
 			}
@@ -74,207 +73,6 @@ func tree(e interface{}) string {
 	}
 }
 
-// sliceExpr holds a syntactic vector to be verified and evaluated.
-type sliceExpr []value.Expr
-
-func (s sliceExpr) Eval(context value.Context) value.Value {
-	v := make([]value.Value, len(s))
-	// Evaluate right to left, as is the usual rule.
-	// This also means things like
-	//	x=1000; x + x=2
-	// (yielding 4) work.
-	for i := len(s) - 1; i >= 0; i-- {
-		v[i] = s[i].Eval(context)
-	}
-	return value.NewVector(v)
-}
-
-var charEscape = map[rune]string{
-	'\\': "\\\\",
-	'\'': "\\'",
-	'\a': "\\a",
-	'\b': "\\b",
-	'\f': "\\f",
-	'\n': "\\n",
-	'\r': "\\r",
-	'\t': "\\t",
-	'\v': "\\v",
-}
-
-func (s sliceExpr) ProgString() string {
-	var b bytes.Buffer
-	// If it's all Char, we can do a prettier job.
-	if s.allChars() {
-		b.WriteRune('\'')
-		for _, v := range s {
-			c := rune(v.(value.Char))
-			esc := charEscape[c]
-			if esc != "" {
-				b.WriteString(esc)
-				continue
-			}
-			if !strconv.IsPrint(c) {
-				if c <= 0xFFFF {
-					fmt.Fprintf(&b, "\\u%04x", c)
-				} else {
-					fmt.Fprintf(&b, "\\U%08x", c)
-				}
-				continue
-			}
-			b.WriteRune(c)
-		}
-		b.WriteRune('\'')
-	} else {
-		for i, v := range s {
-			if i > 0 {
-				b.WriteRune(' ')
-			}
-			if isCompound(v) {
-				b.WriteString("(" + v.ProgString() + ")")
-			} else {
-				b.WriteString(v.ProgString())
-			}
-		}
-	}
-	return b.String()
-}
-
-func (s sliceExpr) allChars() bool {
-	for _, c := range s {
-		if _, ok := c.(value.Char); !ok {
-			return false
-		}
-	}
-	return true
-}
-
-// variableExpr identifies a variable to be looked up and evaluated.
-type variableExpr struct {
-	name  string
-	local int // local index, or 0 for global
-}
-
-func (e *variableExpr) Eval(context value.Context) value.Value {
-	var v value.Value
-	if e.local >= 1 {
-		v = context.Local(e.local)
-	} else {
-		v = context.Global(e.name)
-	}
-	if v == nil {
-		kind := "global"
-		if e.local >= 1 {
-			kind = "local"
-		}
-		value.Errorf("undefined %s variable %q", kind, e.name)
-	}
-	return v
-}
-
-func (e *variableExpr) ProgString() string {
-	return e.name
-}
-
-// isCompound reports whether the item is a non-trivial expression tree, one that
-// may require parentheses around it when printed to maintain correct evaluation order.
-func isCompound(x interface{}) bool {
-	switch x := x.(type) {
-	case value.Char, value.Int, value.BigInt, value.BigRat, value.BigFloat, value.Complex, value.Vector, value.Matrix:
-		return false
-	case sliceExpr, *variableExpr:
-		return false
-	case *index:
-		return isCompound(x.left)
-	default:
-		return true
-	}
-}
-
-type unary struct {
-	op    string
-	right value.Expr
-}
-
-func (u *unary) ProgString() string {
-	return fmt.Sprintf("%s %s", u.op, u.right.ProgString())
-}
-
-func (u *unary) Eval(context value.Context) value.Value {
-	return context.EvalUnary(u.op, u.right.Eval(context).Inner())
-}
-
-type binary struct {
-	op    string
-	left  value.Expr
-	right value.Expr
-}
-
-func (b *binary) ProgString() string {
-	var left string
-	if isCompound(b.left) {
-		left = fmt.Sprintf("(%s)", b.left.ProgString())
-	} else {
-		left = b.left.ProgString()
-	}
-	return fmt.Sprintf("%s %s %s", left, b.op, b.right.ProgString())
-}
-
-func (b *binary) Eval(context value.Context) value.Value {
-	if b.op == "=" {
-		return assignment(context, b)
-	}
-	rhs := b.right.Eval(context).Inner()
-	lhs := b.left.Eval(context)
-	return context.EvalBinary(lhs, b.op, rhs)
-}
-
-type index struct {
-	op    string
-	left  value.Expr
-	right []value.Expr
-}
-
-func (x *index) ProgString() string {
-	var s strings.Builder
-	if isCompound(x.left) {
-		s.WriteString("(")
-		s.WriteString(x.left.ProgString())
-		s.WriteString(")")
-	} else {
-		s.WriteString(x.left.ProgString())
-	}
-	s.WriteString("[")
-	for i, v := range x.right {
-		if i > 0 {
-			s.WriteString("; ")
-		}
-		if v != nil {
-			s.WriteString(v.ProgString())
-		}
-	}
-	s.WriteString("]")
-	return s.String()
-}
-
-func (x *index) Eval(context value.Context) value.Value {
-	return value.Index(context, x, x.left, x.right)
-}
-
-// conditional is a conditional executor: expression ":" expression
-type conditional struct {
-	*binary // Implements Expr through embedding.
-}
-
-var _ = value.Decomposable(conditional{})
-
-func (c conditional) Operator() string {
-	return ":"
-}
-
-func (c conditional) Operands() (left, right value.Expr) {
-	return c.left, c.right
-}
-
 // Parser stores the state for the ivy parser.
 type Parser struct {
 	scanner  *scan.Scanner
@@ -286,7 +84,7 @@ type Parser struct {
 }
 
 // NewParser returns a new parser that will read from the scanner.
-// The context must have have been created by this package's NewContext function.
+// The context must have been created by this package's NewContext function.
 func NewParser(fileName string, scanner *scan.Scanner, context value.Context) *Parser {
 	return &Parser{
 		scanner:  scanner,
@@ -305,6 +103,11 @@ func (p *Parser) Println(args ...interface{}) {
 	fmt.Fprintln(p.context.Config().Output(), args...)
 }
 
+// Print prints the args and writes them to the configured output writer.
+func (p *Parser) Print(args ...interface{}) {
+	fmt.Fprint(p.context.Config().Output(), args...)
+}
+
 func (p *Parser) next() scan.Token {
 	tok := p.peek()
 	if tok.Type != scan.EOF {
@@ -342,6 +145,16 @@ func (p *Parser) errorf(format string, args ...interface{}) {
 	value.Errorf(format, args...)
 }
 
+// source returns the source code spanning the start and end lines.
+func (p *Parser) source(start, end int) string {
+	src := strings.Builder{}
+	for _, s := range p.scanner.History()[start:end] {
+		src.WriteString(s)
+		src.WriteByte('\n')
+	}
+	return src.String()
+}
+
 // Line reads a line of input and returns the values it evaluates.
 // A nil returned slice means there were no values.
 // The boolean reports whether the line is valid.
@@ -349,10 +162,11 @@ func (p *Parser) errorf(format string, args ...interface{}) {
 // Line
 //
 //	) special command '\n'
-//	op function defintion
+//	op function definition
 //	expressionList '\n'
 func (p *Parser) Line() ([]value.Expr, bool) {
 	var ok bool
+	start := len(p.scanner.History()) // Remember this location before any leading comments.
 	if !p.readTokensToNewline(false) {
 		return nil, false
 	}
@@ -365,10 +179,7 @@ func (p *Parser) Line() ([]value.Expr, bool) {
 		p.context.SetConstants()
 		return nil, true
 	case scan.Op:
-		p.functionDefn()
-		return nil, true
-	case scan.OpDelete:
-		p.functionDefn()
+		p.functionDefn(start)
 		return nil, true
 	}
 	exprs, ok := p.expressionList()
@@ -392,11 +203,15 @@ func (p *Parser) readTokensToNewline(inFunction bool) bool {
 			p.errorf("%s", tok)
 		case scan.Newline:
 			// Need a truly blank line to terminate the function body.
-			if !inFunction || len(tok.Text) <= 1 {
+			if !inFunction || len(tok.Text) <= 1 || len(p.tokens) > 0 {
 				return true
 			}
 			continue
 		case scan.EOF:
+			if inFunction && len(p.tokens) == 0 {
+				// EOF is fine for terminating a function.
+				return true
+			}
 			return len(p.tokens) > 0
 		}
 		p.tokens = append(p.tokens, tok)
@@ -417,7 +232,7 @@ func (p *Parser) expressionList() ([]value.Expr, bool) {
 	default:
 		p.errorf("unexpected %s", tok)
 	}
-	if len(exprs) > 0 && p.context.Config().Debug("parse") {
+	if len(exprs) > 0 && p.context.Config().Debug("parse") > 0 {
 		p.Println(tree(exprs))
 	}
 	return exprs, ok
@@ -430,11 +245,11 @@ func (p *Parser) statementList() ([]value.Expr, bool) {
 	expr := p.expr()
 	if expr != nil && p.peek().Type == scan.Colon {
 		tok := p.next()
-		expr = conditional{
-			&binary{
-				left:  expr,
-				op:    tok.Text,
-				right: p.expr(),
+		expr = &value.CondExpr{
+			Cond: &value.BinaryExpr{
+				Left:  expr,
+				Op:    tok.Text,
+				Right: p.expr(),
 			},
 		}
 	}
@@ -458,7 +273,7 @@ func (p *Parser) statementList() ([]value.Expr, bool) {
 //	operand binop expr
 func (p *Parser) expr() value.Expr {
 	tok := p.next()
-	expr := p.operand(tok, true)
+	expr := p.operand(tok)
 	tok = p.peek()
 	switch tok.Type {
 	case scan.EOF, scan.RightParen, scan.RightBrack, scan.Semicolon, scan.Colon:
@@ -466,67 +281,82 @@ func (p *Parser) expr() value.Expr {
 	case scan.Identifier:
 		if p.context.DefinedBinary(tok.Text) {
 			p.next()
-			return &binary{
-				left:  expr,
-				op:    tok.Text,
-				right: p.expr(),
+			return &value.BinaryExpr{
+				Left:  expr,
+				Op:    tok.Text,
+				Right: p.expr(),
 			}
 		}
 	case scan.Assign:
 		p.next()
-		switch lhs := expr.(type) {
-		case *variableExpr, *index:
-			return &binary{
-				left:  lhs,
-				op:    tok.Text,
-				right: p.expr(),
-			}
-		case sliceExpr:
-			for _, v := range lhs {
-				if _, ok := v.(*variableExpr); !ok {
-					p.errorf("cannot assign to %s", v.ProgString())
-				}
-			}
-			return &binary{
-				left:  lhs,
-				op:    tok.Text,
-				right: p.expr(),
-			}
+		p.checkAssign(expr)
+		return &value.BinaryExpr{
+			Left:  expr,
+			Op:    tok.Text,
+			Right: p.expr(),
 		}
-		p.errorf("cannot assign to %s", expr.ProgString())
 	case scan.Operator:
 		p.next()
-		return &binary{
-			left:  expr,
-			op:    tok.Text,
-			right: p.expr(),
+		return &value.BinaryExpr{
+			Left:  expr,
+			Op:    tok.Text,
+			Right: p.expr(),
 		}
 	}
 	p.errorf("after expression: unexpected %s", p.peek())
 	return nil
 }
 
+// checkAssign checks that e is assignable.
+func (p *Parser) checkAssign(e value.Expr) {
+	switch e := e.(type) {
+	default:
+		p.errorf("cannot assign to %s", e.ProgString())
+	case *value.VarExpr:
+		// ok
+	case *value.IndexExpr:
+		switch e.Left.(type) {
+		case *value.VarExpr:
+			// ok
+		case *value.IndexExpr:
+			// Old x[i][j]. Show new syntax.
+			var list []value.Expr
+			var last value.Expr
+			for x := e; x != nil; x, _ = x.Left.(*value.IndexExpr) {
+				list = append(list, x.Right...)
+				last = x.Left
+			}
+			slices.Reverse(list)
+			fixed := &value.IndexExpr{Left: last, Right: list}
+			value.Errorf("cannot assign to %s; use %v", e.ProgString(), fixed.ProgString())
+		}
+	case value.VectorExpr:
+		for _, elem := range e {
+			p.checkAssign(elem)
+		}
+	}
+}
+
 // operand
 //
 //	number
 //	char constant
 //	string constant
 //	vector
-//	operand [ Expr ]...
 //	unop Expr
-func (p *Parser) operand(tok scan.Token, indexOK bool) value.Expr {
+func (p *Parser) operand(tok scan.Token) value.Expr {
 	var expr value.Expr
 	switch tok.Type {
 	case scan.Operator:
-		expr = &unary{
-			op:    tok.Text,
-			right: p.expr(),
+		expr = &value.UnaryExpr{
+			Op:    tok.Text,
+			Right: p.expr(),
 		}
 	case scan.Identifier:
-		if p.context.DefinedUnary(tok.Text) {
-			expr = &unary{
-				op:    tok.Text,
-				right: p.expr(),
+		if p.context.DefinedUnary(strings.Trim(tok.Text, "@")) {
+			expr = &value.UnaryExpr{
+				Op:    tok.Text,
+				Right: p.expr(),
 			}
 			break
 		}
@@ -536,9 +366,6 @@ func (p *Parser) operand(tok scan.Token, indexOK bool) value.Expr {
 	default:
 		p.errorf("unexpected %s", tok)
 	}
-	if indexOK {
-		expr = p.index(expr)
-	}
 	return expr
 }
 
@@ -555,9 +382,9 @@ func (p *Parser) index(expr value.Expr) value.Expr {
 		if tok.Type != scan.RightBrack {
 			p.errorf("expected right bracket, found %s", tok)
 		}
-		expr = &index{
-			left:  expr,
-			right: list,
+		expr = &value.IndexExpr{
+			Left:  expr,
+			Right: list,
 		}
 	}
 	return expr
@@ -596,7 +423,7 @@ func (p *Parser) indexList() []value.Expr {
 //	rational
 //	string
 //	variable
-//	variable '[' Expr ']'
+//	'(' ')'
 //	'(' Expr ')'
 //
 // If the value is a string, value.Expr is nil.
@@ -605,16 +432,21 @@ func (p *Parser) number(tok scan.Token) (expr value.Expr, str string) {
 	text := tok.Text
 	switch tok.Type {
 	case scan.Identifier:
-		expr = p.index(p.variable(text))
+		expr = p.variable(text)
 	case scan.String:
 		str = value.ParseString(text)
 	case scan.Number, scan.Rational, scan.Complex:
 		expr, err = value.Parse(p.context.Config(), text)
 	case scan.LeftParen:
-		expr = p.expr()
-		tok := p.next()
-		if tok.Type != scan.RightParen {
-			p.errorf("expected right paren, found %s", tok)
+		if p.peek().Type == scan.RightParen {
+			p.next()
+			expr = value.VectorExpr{}
+		} else {
+			expr = p.expr()
+			tok := p.next()
+			if tok.Type != scan.RightParen {
+				p.errorf("expected right paren, found %s", tok)
+			}
 		}
 	}
 	if err != nil {
@@ -628,21 +460,22 @@ func (p *Parser) number(tok scan.Token) (expr value.Expr, str string) {
 //
 //	number
 //	string
+//	numberOrVector '[' Expr ']'
 //	numberOrVector...
 func (p *Parser) numberOrVector(tok scan.Token) value.Expr {
 	expr, str := p.number(tok)
 	done := true
 	switch p.peek().Type {
-	case scan.Number, scan.Rational, scan.Complex, scan.String, scan.Identifier, scan.LeftParen:
-		// Further vector elements follow.
+	case scan.Number, scan.Rational, scan.Complex, scan.String, scan.Identifier, scan.LeftParen, scan.LeftBrack:
+		// Further work follows.
 		done = false
 	}
-	var slice sliceExpr
+	var slice value.VectorExpr
 	if expr == nil {
 		// Must be a string.
-		slice = sliceExpr{evalString(str)}
+		slice = value.VectorExpr{p.index(evalString(str))}
 	} else {
-		slice = sliceExpr{expr}
+		slice = value.VectorExpr{p.index(expr)}
 	}
 	if !done {
 	Loop:
@@ -660,13 +493,17 @@ func (p *Parser) numberOrVector(tok scan.Token) value.Expr {
 				expr, str = p.number(p.next())
 				if expr == nil {
 					// Must be a string.
-					slice = append(slice, evalString(str))
-					continue
+					expr = evalString(str)
 				}
 			default:
 				break Loop
 			}
 			slice = append(slice, expr)
+			if p.peek().Type == scan.LeftBrack {
+				// Replace the whole slice so far with the index expression slice[next expression].
+				expr = p.index(slice)
+				slice = append(value.VectorExpr{}, expr)
+			}
 		}
 	}
 	if len(slice) == 1 {
@@ -679,9 +516,9 @@ func isScalar(v value.Value) bool {
 	return v.Rank() == 0
 }
 
-func (p *Parser) variable(name string) *variableExpr {
-	return &variableExpr{
-		name: name,
+func (p *Parser) variable(name string) *value.VarExpr {
+	return &value.VarExpr{
+		Name: name,
 	}
 }
 
@@ -696,5 +533,5 @@ func evalString(str string) value.Expr {
 	for i, c := range r {
 		v[i] = value.Char(c)
 	}
-	return sliceExpr(v)
+	return value.VectorExpr(v)
 }
