@@ -175,31 +175,73 @@ Switch:
 		case "obase":
 			obase = base
 		}
+	case "clear":
+		want := "all"
+		switch p.peek().Text {
+		case "unary", "binary":
+			want = p.next().Text
+		case "var", "vars": // I keep typing "vars" and the resulting error is confusing.
+			p.next()
+			want = "var"
+		}
+		if p.peek().Type == scan.EOF {
+			// Delete every name of specified type.
+			if want == "all" {
+				p.context.UndefineAll(true, true, true)
+			} else {
+				p.context.UndefineAll(want == "unary", want == "binary", want == "var")
+			}
+			break
+		}
+		// Delete only the named items of specified type.
+		for p.peek().Type == scan.Identifier {
+			name := p.next().Text
+			var found bool
+			switch want {
+			case "unary":
+				found = p.context.UndefineOp(name, false)
+			case "binary":
+				found = p.context.UndefineOp(name, true)
+			case "var":
+				found = p.context.UndefineVar(name)
+			case "all":
+				found = p.context.UndefineVar(name)
+				found = p.context.UndefineOp(name, false) || found
+				found = p.context.UndefineOp(name, true) || found
+			}
+			if !found {
+				p.Printf("%q not defined\n", name)
+			}
+		}
+		if p.peek().Type != scan.EOF {
+			p.Println("bad item for clear:", p.peek())
+		}
 	case "cpu":
 		p.Printf("%s\n", conf.PrintCPUTime())
 	case "debug":
 		if p.peek().Type == scan.EOF {
 			for _, f := range config.DebugFlags {
-				p.Printf("%s\t%d\n", f, truth(conf.Debug(f)))
+				p.Printf("%s\t%d\n", f, conf.Debug(f))
 			}
+			p.Println("For trace: 1 traces user-defined only, 2 traces all operators")
 			break Switch
 		}
 		name := p.need(scan.Identifier).Text
+		val := conf.Debug(name)
+		if val < 0 {
+			p.Println("no such debug flag:", name)
+			break Switch
+		}
 		if p.peek().Type == scan.EOF {
 			// Toggle the value
-			if !conf.SetDebug(name, !conf.Debug(name)) {
-				p.Println("no such debug flag:", name)
+			conf.SetDebug(name, truth(val == 0))
+			p.Println(conf.Debug(name))
+		} else {
+			number := p.nextDecimalNumber()
+			if number < 0 {
+				p.Println("illegal value")
 			}
-			if conf.Debug(name) {
-				p.Println("1")
-			} else {
-				p.Println("0")
-			}
-			break
-		}
-		number := p.nextDecimalNumber()
-		if !conf.SetDebug(name, number != 0) {
-			p.Println("no such debug flag:", name)
+			conf.SetDebug(name, number)
 		}
 	case "demo":
 		p.need(scan.EOF)
@@ -221,11 +263,22 @@ Switch:
 		}
 		conf.SetFormat(p.getString())
 	case "get":
-		if p.peek().Type == scan.EOF {
-			p.runFromFile(p.context, defaultFile)
-		} else {
-			p.runFromFile(p.context, p.getString())
+		p.runFromFile(p.context, p.getFile(")get", defaultFile))
+	case "last":
+		history := p.scanner.History()
+		end := len(history) - 1 // -1 to suppress ')last'.
+		start := end - 1
+		if p.peek().Type != scan.EOF {
+			start = end - p.nextDecimalNumber()
 		}
+		// Ignore immediately preceding blank lines.
+		for end > 0 && strings.TrimSpace(history[end-1]) == "" {
+			start--
+			end--
+		}
+		start = max(0, start)
+		end = max(start, end)
+		p.Print(p.source(start, end))
 	case "maxbits":
 		if p.peek().Type == scan.EOF {
 			p.Printf("%d\n", conf.MaxBits())
@@ -277,12 +330,12 @@ Switch:
 		fn := p.context.UnaryFn[name]
 		found := false
 		if fn != nil {
-			p.Println(fn)
+			p.Print(fn.Source)
 			found = true
 		}
 		fn = p.context.BinaryFn[name]
 		if fn != nil {
-			p.Println(fn)
+			p.Print(fn.Source)
 			found = true
 		}
 		if !found {
@@ -318,17 +371,13 @@ Switch:
 	case "save":
 		// Must restore ibase, obase for save.
 		conf.SetBase(ibase, obase)
-		if p.peek().Type == scan.EOF {
-			save(p.context, defaultFile)
-		} else {
-			save(p.context, p.getString())
-		}
+		save(p.context, p.getFile(")save", defaultFile))
 	case "seed":
 		if p.peek().Type == scan.EOF {
 			p.Println(conf.RandomSeed())
 			break Switch
 		}
-		conf.SetRandomSeed(p.nextDecimalNumber64())
+		conf.SetRandomSeed(uint64(p.nextDecimalNumber64()))
 	case "timezone":
 		if p.peek().Type == scan.EOF {
 			_, offset := time.Now().In(conf.Location()).Zone()
@@ -356,10 +405,10 @@ Switch:
 		name := p.need(scan.Identifier).Text
 		value := p.context.Global(name)
 		if value == nil {
-			p.errorf("%q not defined", name)
+			p.errorf("undefined global variable %q", name)
 		}
 		fmt.Printf("%s = ", name)
-		put(conf, conf.Output(), value, false)
+		put(conf, conf.Output(), value.Value(), false)
 		fmt.Print("\n")
 	default:
 		p.errorf(")%s: not recognized", text)
@@ -377,6 +426,29 @@ func (p *Parser) getString() string {
 	return value.ParseString(p.need(scan.String).Text)
 }
 
+// getFile returns the file name on this input line.
+// It might be a quoted string, or else just some text.
+// If nothing is present return the default.
+func (p *Parser) getFile(prefix, def string) string {
+	tok := p.next()
+	switch tok.Type {
+	case scan.EOF:
+		return def
+	case scan.String:
+		return value.ParseString(tok.Text)
+	default:
+		// Just grab the rest of the text on the line.
+		// Must drain the scanner first as we aere cheating it.
+		for p.peek().Type != scan.EOF {
+			p.next()
+		}
+		h := p.scanner.History()
+		line := strings.TrimSpace(h[len(h)-1])
+		line = strings.TrimPrefix(line, prefix)
+		return strings.TrimSpace(line)
+	}
+}
+
 var runDepth = 0
 
 // runFromFile executes the contents of the named file.
@@ -439,11 +511,11 @@ func (p *Parser) runUntilError(name string) error {
 			if val == nil {
 				continue
 			}
-			if _, ok := val.(Assignment); ok {
+			if _, ok := val.(value.QuietValue); ok {
 				continue
 			}
 			p.context.AssignGlobal("_", val)
-			fmt.Fprintf(p.context.Config().Output(), "%v\n", val.Sprint(p.context.Config()))
+			p.Println(val.Sprint(p.context.Config()))
 		}
 		if !ok {
 			return io.EOF
