@@ -40,7 +40,6 @@ const (
 	Number     // simple number
 	Operator   // known operator
 	Op         // "op", operator definition keyword
-	OpDelete   // "opdelete", operator undefinition keyword
 	Rational   // rational number like 2/3
 	Complex    // complex number like 3j2
 	RightBrack // ']'
@@ -82,6 +81,7 @@ type Scanner struct {
 	pos       int    // current position in the input
 	start     int    // start position of this item
 	token     Token
+	history   []string // Record of input.
 }
 
 // loadLine reads the next line of input and stores it in (appends it to) the input.
@@ -99,6 +99,7 @@ func (l *Scanner) loadLine() {
 			l.buf = append(l.buf, c)
 		}
 		if c == '\n' {
+			l.history = append(l.history, string(l.buf[:len(l.buf)-1])) // Drop the newline.
 			break
 		}
 	}
@@ -112,6 +113,11 @@ func (l *Scanner) loadLine() {
 	}
 }
 
+// History returns the history as a slice of strings with newlines elided.
+func (l *Scanner) History() []string {
+	return l.history
+}
+
 // readRune reads the next rune from the input.
 func (l *Scanner) readRune() (rune, int) {
 	if !l.done && l.pos == len(l.input) {
@@ -141,11 +147,17 @@ func (l *Scanner) peek() rune {
 }
 
 // peek2 returns the next two runes ahead, but does not consume anything.
+// If the first rune is a newline, do not read further, and return '\n' for the
+// second rune as well: it shouldn't matter as peek2 is only used for operators,
+// and if we see a newline we don't have an operator.
 func (l *Scanner) peek2() (rune, rune) {
 	pos := l.pos
 	lastWidth := l.lastWidth
 	r1 := l.next()
-	r2 := l.next()
+	r2 := r1
+	if r1 != '\n' {
+		r2 = l.next()
+	}
 	l.pos = pos
 	l.lastWidth = lastWidth
 	return r1, r2
@@ -171,7 +183,7 @@ func (l *Scanner) emit(t Type) stateFn {
 	}
 	text := l.input[l.start:l.pos]
 	config := l.context.Config()
-	if config.Debug("tokens") {
+	if config.Debug("tokens") > 0 {
 		fmt.Fprintf(config.Output(), "%s:%d: emit %s\n", l.name, l.line, Token{t, l.line, text})
 	}
 	l.token = Token{t, l.line, text}
@@ -237,14 +249,10 @@ func lexComment(l *Scanner) stateFn {
 	for {
 		r := l.next()
 		if r == eof || r == '\n' {
+			l.backup()
 			break
 		}
 	}
-	if len(l.input) > 0 {
-		l.pos = len(l.input)
-		// Emitting newline also advances l.line.
-		return l.emit(Newline)
-	}
 	return lexAny
 }
 
@@ -289,12 +297,15 @@ func lexAny(l *Scanner) stateFn {
 			return l.emit(Assign)
 		}
 		l.next()
+		if l.peek() == '=' { // For ===
+			l.next()
+		}
 		fallthrough // for ==
 	case l.isOperator(r):
-		// Must be after after = so == is an operator,
+		// Must be after = so == is an operator,
 		// and after numbers, so '-' can be a sign.
 		return lexOperator
-	case isAlphaNumeric(r):
+	case isAlphaNumeric(r), r == '@':
 		l.backup()
 		return lexIdentifier
 	case r == '[':
@@ -329,37 +340,57 @@ func lexSpace(l *Scanner) stateFn {
 // If the input base is greater than 10, some identifiers
 // are actually numbers. We handle this here.
 func lexIdentifier(l *Scanner) stateFn {
+	eachLeft := 0
+	for l.peek() == '@' {
+		eachLeft++
+		l.next()
+	}
 	for isAlphaNumeric(l.peek()) {
 		l.next()
 	}
 	if !l.atTerminator() {
 		return l.errorf("bad character %#U", l.next())
 	}
+	eachRight := 0
+	for l.peek() == '@' {
+		eachRight++
+		l.next()
+	}
 	// Some identifiers are operators.
-	word := l.input[l.start:l.pos]
+	word := l.input[l.start+eachLeft : l.pos-eachRight]
 	switch {
 	case word == "op":
 		return l.emit(Op)
-	case word == "opdelete":
-		return l.emit(OpDelete)
 	case word == "o" && l.peek() == '.':
 		return lexOperator
 	case l.defined(word):
 		return lexOperator
+	case word == "" || eachLeft > 0:
+		return l.errorf("invalid operator %q", l.input[l.start:l.pos])
 	case isAllDigits(word, l.context.Config().InputBase()):
 		// Mistake: back up and scan it as a number.
 		l.pos = l.start
 		return lexComplex
 	}
+	if !l.context.UserDefined(word, false) {
+		// Lexed word@ but word is not a unary operator; leave @ for next time.
+		l.pos -= eachRight
+	}
 	return l.emit(Identifier)
 }
 
 // lexOperator completes scanning an operator. We have already accepted the + or
 // whatever; there may be a reduction or inner or outer product.
 func lexOperator(l *Scanner) stateFn {
+	// Might have been called for symbol like + and not scanned trailing @s yet.
+	for l.peek() == '@' {
+		l.next()
+	}
+
 	// It might be an inner product or reduction, but only if it is a binary operator.
 	word := l.input[l.start:l.pos]
-	if word == "o" || value.BinaryOps[word] != nil || l.context.UserDefined(word, true) {
+	w := strings.Trim(word, "@")
+	if word == "o" || value.BinaryOps[w] != nil || l.context.UserDefined(w, true) {
 		switch l.peek() {
 		case '/', '\\':
 			// Reduction or scan.
@@ -377,7 +408,7 @@ func lexOperator(l *Scanner) stateFn {
 			prevPos := l.pos
 			r := l.next()
 			switch {
-			case l.isOperator(r):
+			case l.isOperatorToken(r):
 			case isAlphaNumeric(r):
 				for isAlphaNumeric(r) {
 					r = l.next()
@@ -456,7 +487,11 @@ func acceptNumber(l *Scanner, realPart bool) (bool, stateFn) {
 		}
 	}
 	if !l.scanNumber(true, realPart) {
-		return false, l.errorf("bad number syntax: %s", l.input[l.start:l.pos])
+		// We could error out here, but that prevents the )get special
+		// command from handling a dot in a file name, so we return what we
+		// have and let the number syntax checking catch the error
+		// otherwise.
+		return true, l.emit(Number)
 	}
 	r := l.peek()
 	if r != '/' {
@@ -486,10 +521,17 @@ func acceptNumber(l *Scanner, realPart bool) (bool, stateFn) {
 func (l *Scanner) scanNumber(followingSlashOK, followingJOK bool) bool {
 	base := l.context.Config().InputBase()
 	digits := digitsForBase(base)
-	// If base 0, acccept octal for 0 or hex for 0x or 0X.
+	// If base 0, accept octal for 0 or hex for 0x or 0X.
 	if base == 0 {
-		if l.accept("0") && l.accept("xX") {
-			digits = digitsForBase(16)
+		if l.accept("0") {
+			switch {
+			case l.accept("xX"):
+				digits = digitsForBase(16)
+			case l.accept("oO"):
+				digits = digitsForBase(8)
+			case l.accept("bB"):
+				digits = digitsForBase(2)
+			}
 		}
 		// Otherwise leave it decimal (0); strconv.ParseInt will take care of it.
 		// We can't set it to 8 in case it's a leading-0 float like 0.69 or 09e4.
@@ -498,7 +540,7 @@ func (l *Scanner) scanNumber(followingSlashOK, followingJOK bool) bool {
 	if l.accept(".") {
 		l.acceptRun(digits)
 	}
-	if l.accept("eE") {
+	if l.accept("eEpP") {
 		l.accept("+-")
 		l.acceptRun("0123456789")
 	}
@@ -524,12 +566,14 @@ func (l *Scanner) scanNumber(followingSlashOK, followingJOK bool) bool {
 var digits [36 + 1]string // base 36 is OK.
 
 const (
-	decimal = "0123456789"
-	lower   = "abcdefghijklmnopqrstuvwxyz"
-	upper   = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+	decimal = "_0123456789"
+	lower   = "_abcdefghijklmnopqrstuvwxyz"
+	upper   = "_ABCDEFGHIJKLMNOPQRSTUVWXYZ"
 )
 
 // digitsForBase returns the digit set for numbers in the specified base.
+// Underscores are always admitted, but will only be accepted during
+// parsing (by strconv.ParseInt) if ibase is 0. This is consistent with Go.
 func digitsForBase(base int) string {
 	if base == 0 {
 		base = 10
@@ -541,9 +585,9 @@ func digitsForBase(base int) string {
 			// Whatever the input base, if it's <= 10 let the parser
 			// decide if it's valid. This also helps us get the always-
 			// base-10 numbers for )specials.
-			d = decimal[:10]
+			d = decimal[:1+10]
 		} else {
-			d = decimal + lower[:base-10] + upper[:base-10]
+			d = decimal + lower[:1+base-10] + upper[:1+base-10]
 		}
 		digits[base] = d
 	}
@@ -669,9 +713,25 @@ func isAllDigits(s string, base int) bool {
 	return true
 }
 
-// isOperator reports whether r is an operator. It may advance the lexer one character
-// if it is a two-character operator.
+// isOperator reports whether r is an operator. It may advance the lexer
+// if it is a multi-character operator.
 func (l *Scanner) isOperator(r rune) bool {
+	n := 0
+	for r == '@' {
+		r = l.next()
+		n++
+	}
+	if !l.isOperatorToken(r) {
+		l.pos -= n
+		return false
+	}
+	for l.peek() == '@' {
+		l.next()
+	}
+	return true
+}
+
+func (l *Scanner) isOperatorToken(r rune) bool {
 	switch r {
 	case '?', '+', '-', '/', '&', '|', '^':
 		// No follow-on possible.
@@ -679,9 +739,12 @@ func (l *Scanner) isOperator(r rune) bool {
 		if l.peek() == '%' {
 			l.next()
 		}
-	case '!':
+	case '!', '=':
 		if l.peek() == '=' {
 			l.next()
+			if l.peek() == '=' { // For !== and ===
+				l.next()
+			}
 		}
 	case '>':
 		switch l.peek() {
@@ -697,11 +760,6 @@ func (l *Scanner) isOperator(r rune) bool {
 		if l.peek() == '*' {
 			l.next()
 		}
-	case '=':
-		if l.peek() != '=' {
-			return false
-		}
-		l.next()
 	default:
 		return false
 	}
