@@ -177,18 +177,16 @@ func toBool(t Value) bool {
 	panic("not reached")
 }
 
-// andBool is like toBool but handles vectors by and'ing the values together.
-// The results are known to be Ints, as they come from comparison operations.
-func andBool(t Value) bool {
-	if v, ok := t.(Vector); ok {
-		for _, x := range v {
-			if x == zero {
-				return false
-			}
+// allEqual returns an elementwise boolean AND of the two slices
+// using OrderedCompare on the individual elements. It will exit
+// early if a false entry is encountered.
+func allEqual(c Context, x *Vector, xlo int, y *Vector, ylo, n int) bool {
+	for i := range n {
+		if OrderedCompare(c, x.At(xlo+i), y.At(ylo+i)) != 0 {
+			return false
 		}
-		return true
 	}
-	return t.(Int) == one
+	return true
 }
 
 var BinaryOps = make(map[string]BinaryOp)
@@ -201,16 +199,16 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
-					return NewComplex(u, v)
+					return NewComplex(u, v).shrink()
 				},
 				bigIntType: func(c Context, u, v Value) Value {
-					return NewComplex(u, v)
+					return NewComplex(u, v).shrink()
 				},
 				bigRatType: func(c Context, u, v Value) Value {
-					return NewComplex(u, v)
+					return NewComplex(u, v).shrink()
 				},
 				bigFloatType: func(c Context, u, v Value) Value {
-					return NewComplex(u, v)
+					return NewComplex(u, v).shrink()
 				},
 			},
 		},
@@ -221,6 +219,14 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
+					// Avoid Int->Value interface allocation;
+					// especially effective for sparse matrices.
+					if u.(Int) == 0 {
+						return v
+					}
+					if v.(Int) == 0 {
+						return u
+					}
 					return (u.(Int) + v.(Int)).maybeBig()
 				},
 				bigIntType: func(c Context, u, v Value) Value {
@@ -235,7 +241,7 @@ func init() {
 					return binaryBigFloatOp(c, u, (*big.Float).Add, v)
 				},
 				complexType: func(c Context, u, v Value) Value {
-					return u.(Complex).add(c, v.(Complex))
+					return u.(Complex).add(c, v.(Complex)).shrink()
 				},
 			},
 		},
@@ -246,6 +252,11 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
+					// Avoid Int->Value interface allocation;
+					// especially effective for sparse matrices.
+					if v.(Int) == 0 {
+						return u
+					}
 					return (u.(Int) - v.(Int)).maybeBig()
 				},
 				bigIntType: func(c Context, u, v Value) Value {
@@ -260,7 +271,7 @@ func init() {
 					return binaryBigFloatOp(c, u, (*big.Float).Sub, v)
 				},
 				complexType: func(c Context, u, v Value) Value {
-					return u.(Complex).sub(c, v.(Complex))
+					return u.(Complex).sub(c, v.(Complex)).shrink()
 				},
 			},
 		},
@@ -271,6 +282,14 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
+					// Avoid Int->Value interface allocation;
+					// especially effective for sparse matrices.
+					if u.(Int) == 1 || v.(Int) == 0 {
+						return v
+					}
+					if v.(Int) == 1 || u.(Int) == 0 {
+						return u
+					}
 					return (u.(Int) * v.(Int)).maybeBig()
 				},
 				bigIntType: func(c Context, u, v Value) Value {
@@ -284,7 +303,7 @@ func init() {
 					return binaryBigFloatOp(c, u, (*big.Float).Mul, v)
 				},
 				complexType: func(c Context, u, v Value) Value {
-					return u.(Complex).mul(c, v.(Complex))
+					return u.(Complex).mul(c, v.(Complex)).shrink()
 				},
 			},
 		},
@@ -304,7 +323,7 @@ func init() {
 					return binaryBigFloatOp(c, u, (*big.Float).Quo, v)
 				},
 				complexType: func(c Context, u, v Value) Value {
-					return u.(Complex).div(c, v.(Complex))
+					return u.(Complex).div(c, v.(Complex)).shrink()
 				},
 			},
 		},
@@ -362,12 +381,12 @@ func init() {
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
 					// Projection of u onto v
-					uu, vv := u.(Vector), v.(Vector)
-					if len(uu) != len(vv) {
-						Errorf("mismatched lengths %d, %d in vector mdiv", len(uu), len(vv))
+					uu, vv := u.(*Vector), v.(*Vector)
+					if uu.Len() != vv.Len() {
+						Errorf("mismatched lengths %d, %d in vector mdiv", uu.Len(), vv.Len())
 					}
-					for i := range uu {
-						if whichType(uu[i]) >= complexType || whichType(vv[i]) >= complexType {
+					for i := range uu.All() {
+						if whichType(uu.At(i)) >= complexType || whichType(vv.At(i)) >= complexType {
 							Errorf("non-real element in vector mdiv")
 						}
 					}
@@ -515,6 +534,14 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
+					// Avoid Int->Value interface allocation;
+					// especially effective for sparse matrices.
+					if u.(Int) == 0 || v.(Int) == -1 {
+						return u
+					}
+					if v.(Int) == 0 || u.(Int) == -1 {
+						return v
+					}
 					return u.(Int) & v.(Int)
 				},
 				bigIntType: func(c Context, u, v Value) Value {
@@ -529,6 +556,14 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
+					// Avoid Int->Value interface allocation;
+					// especially effective for sparse matrices.
+					if u.(Int) == 0 || v.(Int) == -1 {
+						return v
+					}
+					if v.(Int) == 0 || u.(Int) == -1 {
+						return u
+					}
 					return u.(Int) | v.(Int)
 				},
 				bigIntType: func(c Context, u, v Value) Value {
@@ -543,6 +578,14 @@ func init() {
 			whichType:   binaryArithType,
 			fn: [numType]binaryFn{
 				intType: func(c Context, u, v Value) Value {
+					// Avoid Int->Value interface allocation;
+					// especially effective for sparse matrices.
+					if u.(Int) == 0 {
+						return v
+					}
+					if v.(Int) == 0 {
+						return u
+					}
 					return u.(Int) ^ v.(Int)
 				},
 				bigIntType: func(c Context, u, v Value) Value {
@@ -895,13 +938,53 @@ func init() {
 					if A > B {
 						Errorf("left operand larger than right in %d?%d", A, B)
 					}
-					ints := c.Config().Random().Perm(int(B))
 					origin := c.Config().Origin()
-					res := make([]Value, A)
-					for i := range res {
-						res[i] = Int(ints[i] + origin)
+					res := newVectorEditor(int(A), nil)
+					r := c.Config().Random()
+					// If B is large and A is a significant fraction of B, use Perm. Otherwise
+					// just guess and avoid duplicates because if B is large,
+					// even 1?B can be expensive.
+					if B < 100 || A*5 > B {
+						ints := r.Perm(int(B))
+						for i := range res.Len() {
+							res.Set(i, Int(ints[i]+origin))
+						}
+					} else {
+						set := make(map[int64]bool)
+						for i := range int(A) {
+							for {
+								x := r.Int64N(int64(B))
+								if !set[x] {
+									set[x] = true
+									res.Set(i, Int(x)+Int(origin))
+									break
+								}
+							}
+						}
 					}
-					return NewVector(res)
+					return res.Publish()
+				},
+				bigIntType: func(c Context, u, v Value) Value {
+					// The count, must be modest. A million big ints is enough.
+					if !u.(BigInt).IsInt64() || u.(BigInt).Int64() > 1e6 {
+						Errorf("negative or too-large operand in %d?%d", u, v)
+					}
+					A := int(u.(BigInt).Int64())
+					B := v.(BigInt)
+					res := newVectorEditor(A, nil)
+					set := make(map[string]bool) // Easiest way to make a key from a big.Int.
+					for i := range A {
+						for {
+							x := bigIntRand(c, big.NewInt(0), B.Int)
+							s := x.String()
+							if !set[s] {
+								set[s] = true
+								res.Set(i, BigInt{x})
+								break
+							}
+						}
+					}
+					return res.Publish()
 				},
 			},
 		},
@@ -911,28 +994,28 @@ func init() {
 			whichType: vectorAndAtLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					// A decode B is the result of polyomial B at x=A.
+					// A decode B is the result of polynomial B at x=A.
 					// If A is a vector, the elements of A align with B.
-					A, B := u.(Vector), v.(Vector)
-					if len(A) == 0 || len(B) == 0 {
+					A, B := u.(*Vector), v.(*Vector)
+					if A.Len() == 0 || B.Len() == 0 {
 						return zero
 					}
-					if allChars(A) {
+					if A.AllChars() {
 						// Special case for times.
 						return decodeTime(c, A, B)
 					}
-					if len(A) == 1 || len(B) == 1 || len(A) == len(B) {
+					if A.Len() == 1 || B.Len() == 1 || A.Len() == B.Len() {
 						result := Value(zero)
 						prod := Value(one)
-						get := func(v Vector, i int) Value {
-							if len(v) == 1 {
-								return v[0]
+						get := func(v *Vector, i int) Value {
+							if v.Len() == 1 {
+								return v.At(0)
 							}
-							return v[i]
+							return v.At(i)
 						}
-						n := len(A)
-						if len(B) > n {
-							n = len(B)
+						n := A.Len()
+						if B.Len() > n {
+							n = B.Len()
 						}
 						for i := n - 1; i >= 0; i-- {
 							result = c.EvalBinary(result, "+", c.EvalBinary(prod, "*", get(B, i)))
@@ -940,47 +1023,47 @@ func init() {
 						}
 						return result
 					}
-					if len(A) != len(B) {
+					if A.Len() != B.Len() {
 						Errorf("decode of unequal lengths")
 					}
 					return nil
 				},
 				matrixType: func(c Context, u, v Value) Value {
-					A, B := u.(Vector), v.(*Matrix)
-					if len(A) != 1 && B.shape[0] != 1 && len(A) != B.shape[0] {
-						Errorf("decode of length %d and shape %s", len(A), NewIntVector(B.shape...))
+					A, B := u.(*Vector), v.(*Matrix)
+					if A.Len() != 1 && B.shape[0] != 1 && A.Len() != B.shape[0] {
+						Errorf("decode of length %d and shape %s", A.Len(), NewIntVector(B.shape...))
 					}
 					shape := B.shape[1:]
-					elems := make([]Value, len(B.data)/B.shape[0])
-					get := func(v Vector, i int) Value {
-						if len(v) == 1 {
-							return v[0]
+					elems := newVectorEditor(B.data.Len()/B.shape[0], nil)
+					get := func(v *Vector, i int) Value {
+						if v.Len() == 1 {
+							return v.At(0)
 						}
-						return v[i]
+						return v.At(i)
 					}
-					n := len(A)
+					n := A.Len()
 					if B.shape[0] > n {
 						n = B.shape[0]
 					}
-					pfor(true, n, len(elems), func(lo, hi int) {
+					pfor(true, n, elems.Len(), func(lo, hi int) {
 						for j := lo; j < hi; j++ {
 							result := Value(zero)
 							prod := Value(one)
 							for i := n - 1; i >= 0; i-- {
-								Bslice := B.data
+								bj := j
 								if B.shape[0] > 1 {
-									Bslice = B.data[i*len(elems) : (i+1)*len(elems)]
+									bj += i * elems.Len()
 								}
-								result = c.EvalBinary(result, "+", c.EvalBinary(prod, "*", Bslice[j]))
+								result = c.EvalBinary(result, "+", c.EvalBinary(prod, "*", B.data.At(bj)))
 								prod = c.EvalBinary(prod, "*", get(A, i))
 							}
-							elems[j] = result
+							elems.Set(j, result)
 						}
 					})
 					if len(shape) == 1 {
-						return NewVector(elems)
+						return elems.Publish()
 					}
-					return NewMatrix(shape, elems)
+					return NewMatrix(shape, elems.Publish())
 				},
 			},
 		},
@@ -999,72 +1082,72 @@ func init() {
 					// If they are negative the answers disagree with APL because
 					// of how modulo arithmetic works.
 					const op = "encode"
-					A, B := u.(Vector), v.(Vector)
-					if allChars(A) {
+					A, B := u.(*Vector), v.(*Vector)
+					if A.AllChars() {
 						// Special case for times.
 						return encodeTime(c, A, B)
 					}
 					// Scalar.
-					if len(A) == 1 && len(B) == 1 {
-						_, rem := QuoRem(op, c, B[0], A[0])
+					if A.Len() == 1 && B.Len() == 1 {
+						_, rem := QuoRem(op, c, B.At(0), A.At(0))
 						return rem
 					}
 					// Vector.
-					if len(B) == 1 {
+					if B.Len() == 1 {
 						// 2 2 2 2 encode 11 is 1 0 1 1.
-						elems := make([]Value, len(A))
-						b := B[0]
-						for i := len(A) - 1; i >= 0; i-- {
-							quo, rem := QuoRem(op, c, b, A[i])
-							elems[i] = rem
+						elems := newVectorEditor(A.Len(), nil)
+						b := B.At(0)
+						for i := A.Len() - 1; i >= 0; i-- {
+							quo, rem := QuoRem(op, c, b, A.At(i))
+							elems.Set(i, rem)
 							b = quo
 						}
-						return NewVector(elems)
+						return elems.Publish()
 					}
-					if len(A) == 1 {
+					if A.Len() == 1 {
 						// 3 encode 1 2 3 4 is 1 2 0 1
-						elems := make([]Value, len(B))
-						a := A[0]
-						for i := range B {
-							_, rem := QuoRem(op, c, B[i], a)
-							elems[i] = rem
+						elems := newVectorEditor(B.Len(), nil)
+						a := A.At(0)
+						for i := range B.All() {
+							_, rem := QuoRem(op, c, B.At(i), a)
+							elems.Set(i, rem)
 						}
-						return NewVector(elems)
+						return elems.Publish()
 					}
 					// Matrix.
 					// 2 2 encode 1 2 3 has 3 columns encoding 1 2 3 downwards:
 					// 0 1 1
 					// 1 0 1
-					elems := make([]Value, len(A)*len(B))
-					shape := []int{len(A), len(B)}
-					pfor(true, len(A), len(B), func(lo, hi int) {
+					elems := newVectorEditor(A.Len()*B.Len(), nil)
+					shape := []int{A.Len(), B.Len()}
+					pfor(true, A.Len(), B.Len(), func(lo, hi int) {
 						for j := lo; j < hi; j++ {
-							b := B[j]
-							for i := len(A) - 1; i >= 0; i-- {
-								quo, rem := QuoRem(op, c, b, A[i])
-								elems[j+i*len(B)] = rem
+							b := B.At(j)
+							for i := A.Len() - 1; i >= 0; i-- {
+								quo, rem := QuoRem(op, c, b, A.At(i))
+								elems.Set(j+i*B.Len(), rem)
 								b = quo
 							}
 						}
 					})
-					return NewMatrix(shape, elems)
+					return NewMatrix(shape, elems.Publish())
 				},
 				matrixType: func(c Context, u, v Value) Value {
-					A, B := u.(Vector), v.(*Matrix)
-					elems := make([]Value, len(A)*len(B.data))
-					shape := append([]int{len(A)}, B.Shape()...)
+					A, B := u.(*Vector), v.(*Matrix)
+					elems := newVectorEditor(A.Len()*B.data.Len(), nil)
+					shape := append([]int{A.Len()}, B.Shape()...)
 					const op = "encode"
-					pfor(true, len(A), len(B.data), func(lo, hi int) {
+					pfor(true, A.Len(), B.data.Len(), func(lo, hi int) {
 						for j := lo; j < hi; j++ {
-							b := B.data[j]
-							for i := len(A) - 1; i >= 0; i-- {
-								quo, rem := QuoRem(op, c, b, A[i])
-								elems[j+i*len(B.data)] = rem
+							b := B.data.At(j)
+							for i := A.Len() - 1; i >= 0; i-- {
+								quo, rem := QuoRem(op, c, b, A.At(i))
+								elems.Set(j+i*B.data.Len(), rem)
 								b = quo
 							}
 						}
 					})
-					return NewMatrix(shape, elems)
+					return NewMatrix(shape, elems.Publish())
 				},
 			},
 		},
@@ -1075,13 +1158,13 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					return NewVector(membership(c, u.(Vector), v.(Vector))).shrink()
+					return membership(c, u.(*Vector), v.(*Vector)).shrink()
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					m := u.(*Matrix)
 					data := membership(c, m.data, v.(*Matrix).data)
 					if m.Rank() <= 1 {
-						return NewVector(data).shrink()
+						return data.shrink()
 					}
 					return NewMatrix(m.shape, data)
 				},
@@ -1094,34 +1177,34 @@ func init() {
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
 					// A⍳B: The location (index) of B in A; 0 if not found. (APL does 1+⌈/⍳⍴A)
-					A, B := u.(Vector), v.(Vector)
+					A, B := u.(*Vector), v.(*Vector)
 					type indexed struct {
 						v     Value
 						index int
 					}
 					origin := c.Config().Origin()
-					sortedA := make([]indexed, len(A))
-					for i, a := range A {
+					sortedA := make([]indexed, A.Len())
+					for i, a := range A.All() {
 						sortedA[i] = indexed{a, i + origin}
 					}
 					sort.SliceStable(sortedA, func(i, j int) bool {
-						return c.EvalBinary(sortedA[i].v, "<", sortedA[j].v) == one
+						return OrderedCompare(c, sortedA[i].v, sortedA[j].v) < 0
 					})
-					indices := make([]Value, len(B))
-					work := 2 * (1 + int(math.Log2(float64(len(A)))))
-					pfor(true, work, len(B), func(lo, hi int) {
+					indices := newVectorEditor(B.Len(), nil)
+					work := 2 * (1 + int(math.Log2(float64(A.Len()))))
+					pfor(true, work, B.Len(), func(lo, hi int) {
 						for i := lo; i < hi; i++ {
-							b := B[i]
-							indices[i] = Int(origin - 1)
+							b := B.At(i)
+							indices.Set(i, Int(origin-1))
 							pos := sort.Search(len(sortedA), func(j int) bool {
-								return c.EvalBinary(sortedA[j].v, ">=", b) == one
+								return OrderedCompare(c, sortedA[j].v, b) >= 0
 							})
-							if pos < len(sortedA) && c.EvalBinary(sortedA[pos].v, "==", b) == one {
-								indices[i] = Int(sortedA[pos].index)
+							if pos < len(sortedA) && OrderedCompare(c, sortedA[pos].v, b) == 0 {
+								indices.Set(i, Int(sortedA[pos].index))
 							}
 						}
 					})
-					return NewVector(indices)
+					return indices.Publish()
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					A, B := u.(*Matrix), v.(*Matrix)
@@ -1135,23 +1218,23 @@ func init() {
 					if len(shape) == 0 {
 						shape = []int{1}
 					}
-					n := len(A.data) / A.shape[0] // elements in each comparison
-					indices := make([]Value, len(B.data)/n)
-					pfor(true, n, len(B.data)/n, func(lo, hi int) {
+					n := A.data.Len() / A.shape[0] // elements in each comparison
+					indices := newVectorEditor(B.data.Len()/n, nil)
+					pfor(true, n, B.data.Len()/n, func(lo, hi int) {
 						for i := lo; i < hi; i++ {
-							indices[i] = Int(origin - 1)
-							for j := 0; j < len(A.data); j += n {
-								if andBool(c.EvalBinary(A.data[j:j+n], "==", B.data[i*n:(i+1)*n])) {
-									indices[i] = Int(j/n + origin)
+							indices.Set(i, Int(origin-1))
+							for j := 0; j < A.data.Len(); j += n {
+								if allEqual(c, A.data, j, B.data, i*n, n) {
+									indices.Set(i, Int(j/n+origin))
 									break
 								}
 							}
 						}
 					})
 					if len(shape) == 1 {
-						return NewVector(indices)
+						return indices.Publish()
 					}
-					return NewMatrix(shape, indices)
+					return NewMatrix(shape, indices.Publish())
 				},
 			},
 		},
@@ -1243,7 +1326,7 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					return reshape(u.(Vector), v.(Vector))
+					return reshape(u.(*Vector), v.(*Vector))
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					// LHS must be a vector underneath.
@@ -1261,9 +1344,7 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					uu := u.(Vector)
-					uu = uu[:len(uu):len(uu)]
-					return append(uu, v.(Vector)...)
+					return u.(*Vector).catenate(v.(*Vector))
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					return u.(*Matrix).catenate(v.(*Matrix))
@@ -1276,9 +1357,8 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					uu := u.(Vector)
-					uu = uu[:len(uu):len(uu)]
-					return append(uu, v.(Vector)...)
+					uu := u.(*Vector)
+					return NewVectorSeq(uu.All(), v.(*Vector).All())
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					return u.(*Matrix).catenateFirst(v.(*Matrix))
@@ -1291,46 +1371,37 @@ func init() {
 			whichType: vectorAndAtLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					uu := u.(Vector)
-					vv := v.(Vector)
-					if len(uu) != 1 {
+					uu := u.(*Vector)
+					vv := v.(*Vector)
+					if uu.Len() != 1 {
 						// Need to expand to a matrix.
-						return NewMatrix([]int{len(vv)}, vv).take(c, uu)
+						return NewMatrix([]int{vv.Len()}, vv).take(c, uu)
 					}
-					n, ok := uu[0].(Int) // Number of elements in result.
-					if !ok {
-						Errorf("bad count %s in take", uu[0])
-					}
-					len := Int(len(vv)) // Length of rhs vector.
+					n := uu.intAt(0, "take count") // Number of elements in result.
+					len := vv.Len()                // Length of rhs vector.
 					nElems := n
 					if n < 0 {
 						nElems = -nElems
 					}
-					elems := make([]Value, nElems)
 					fill := vv.fillValue()
 					switch {
 					case n < 0:
 						if nElems > len {
-							for i := 0; i < int(nElems-len); i++ {
-								elems[i] = fill
-							}
-							copy(elems[nElems-len:], vv)
-						} else {
-							copy(elems, vv[len-nElems:])
+							return NewVectorSeq(repeat(fill, nElems-len), vv.All())
 						}
+						return NewVectorSeq(vv.Slice(int(len-nElems), vv.Len()))
 					case n == 0:
+						return NewVectorSeq()
 					case n > 0:
 						if nElems > len {
-							for i := len; i < nElems; i++ {
-								elems[i] = fill
-							}
+							return NewVectorSeq(vv.All(), repeat(fill, int(nElems-len)))
 						}
-						copy(elems, vv)
+						return NewVectorSeq(vv.Slice(0, int(nElems)))
 					}
-					return NewVector(elems)
+					panic("unreachable")
 				},
 				matrixType: func(c Context, u, v Value) Value {
-					return v.(*Matrix).take(c, u.(Vector))
+					return v.(*Matrix).take(c, u.(*Vector))
 				},
 			},
 		},
@@ -1340,33 +1411,30 @@ func init() {
 			whichType: vectorAndAtLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					vv := v.(Vector)
-					uu, ok := u.(Vector)
-					if !ok || len(uu) != 1 {
+					vv := v.(*Vector)
+					uu, ok := u.(*Vector)
+					if !ok || uu.Len() != 1 {
 						Errorf("bad count %s in drop", u)
 					}
-					n, ok := uu[0].(Int) // Number of elements in result.
-					if !ok {
-						Errorf("bad count %s in drop", uu[0])
-					}
-					len := Int(len(vv)) // Length of rhs vector.
+					n := uu.intAt(0, "drop count")
+					len := vv.Len() // Length of rhs vector.
 					switch {
 					case n < 0:
 						if -n > len {
 							return empty
 						}
-						vv = vv[0 : len+n]
+						vv = NewVectorSeq(vv.Slice(0, len+n))
 					case n == 0:
 					case n > 0:
 						if n > len {
 							return empty
 						}
-						vv = vv[n:]
+						vv = NewVectorSeq(vv.Slice(n, vv.Len()))
 					}
-					return vv.Copy()
+					return vv
 				},
 				matrixType: func(c Context, u, v Value) Value {
-					return v.(*Matrix).drop(c, u.(Vector))
+					return v.(*Matrix).drop(c, u.(*Vector))
 				},
 			},
 		},
@@ -1376,23 +1444,18 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					countVec := u.(Vector)
-					count, ok := countVec[0].(Int)
-					if !ok {
+					countVec := u.(*Vector)
+					if countVec.Len() != 1 {
 						Errorf("rot: count must be small integer")
 					}
-					return v.(Vector).rotate(int(count))
+					return v.(*Vector).rotate(countVec.intAt(0, "rot count"))
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					countMat := u.(*Matrix)
-					if countMat.Rank() != 1 || len(countMat.data) != 1 {
+					if countMat.Rank() != 1 || countMat.data.Len() != 1 {
 						Errorf("rot: count must be small integer")
 					}
-					count, ok := countMat.data[0].(Int)
-					if !ok {
-						Errorf("rot: count must be small integer")
-					}
-					return v.(*Matrix).rotate(int(count))
+					return v.(*Matrix).rotate(countMat.data.intAt(0, "rot count"))
 				},
 			},
 		},
@@ -1402,26 +1465,18 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					countVec := u.(Vector)
-					if len(countVec) != 1 {
+					countVec := u.(*Vector)
+					if countVec.Len() != 1 {
 						Errorf("flip: count must be small integer")
 					}
-					count, ok := countVec[0].(Int)
-					if !ok {
-						Errorf("flip: count must be small integer")
-					}
-					return v.(Vector).rotate(int(count))
+					return v.(*Vector).rotate(countVec.intAt(0, "flip count"))
 				},
 				matrixType: func(c Context, u, v Value) Value {
 					countMat := u.(*Matrix)
-					if countMat.Rank() != 1 || len(countMat.data) != 1 {
+					if countMat.Rank() != 1 || countMat.data.Len() != 1 {
 						Errorf("flip: count must be small integer")
 					}
-					count, ok := countMat.data[0].(Int)
-					if !ok {
-						Errorf("flip: count must be small integer")
-					}
-					return v.(*Matrix).vrotate(int(count))
+					return v.(*Matrix).vrotate(countMat.data.intAt(0, "flip count"))
 				},
 			},
 		},
@@ -1431,15 +1486,15 @@ func init() {
 			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					i := u.(Vector)
-					j := v.(Vector)
-					if len(i) == 0 {
+					i := u.(*Vector)
+					j := v.(*Vector)
+					if i.Len() == 0 {
 						return empty
 					}
 					// All lhs values must be small integers.
 					var count int64
 					numLeft := 0
-					for _, x := range i {
+					for _, x := range i.All() {
 						y, ok := x.(Int)
 						if !ok {
 							Errorf("fill: left operand must be small integers")
@@ -1454,13 +1509,13 @@ func init() {
 							count += int64(y)
 						}
 					}
-					if numLeft != len(j) {
-						Errorf("fill: count > 0 on left (%d) must equal length of right (%d)", numLeft, len(j))
+					if numLeft != j.Len() {
+						Errorf("fill: count > 0 on left (%d) must equal length of right (%d)", numLeft, j.Len())
 					}
 					if count > 1e8 {
 						Errorf("fill: result too large: %d elements", count)
 					}
-					result := make([]Value, 0, count)
+					result := newVectorEditor(0, nil)
 					jx := 0
 					var zeroVal Value
 					if j.AllChars() {
@@ -1468,80 +1523,58 @@ func init() {
 					} else {
 						zeroVal = zero
 					}
-					for _, x := range i {
+					for _, x := range i.All() {
 						y := x.(Int)
 						switch {
 						case y == 0:
-							result = append(result, zeroVal)
+							result.Append(zeroVal)
 						case y < 0:
 							for y = -y; y > 0; y-- {
-								result = append(result, zeroVal)
+								result.Append(zeroVal)
 							}
 						default:
 							for ; y > 0; y-- {
-								result = append(result, j[jx])
+								result.Append(j.At(jx))
 							}
 							jx++
 						}
 					}
-					return NewVector(result)
+					return result.Publish()
 				},
 			},
 		},
 
 		{
 			name:      "sel",
-			whichType: vectorAndAtLeastVectorType,
+			whichType: atLeastVectorType,
 			fn: [numType]binaryFn{
 				vectorType: func(c Context, u, v Value) Value {
-					i := u.(Vector)
-					j := v.(Vector)
-					if len(i) == 0 {
-						return empty
-					}
-					// All lhs values must be small integers.
-					var count int64
-					for _, x := range i {
-						y, ok := x.(Int)
-						if !ok {
-							Errorf("sel: left operand must be small integers")
-						}
-						if y < 0 {
-							count -= int64(y)
-						} else {
-							count += int64(y)
-						}
-					}
-					if count > 1e8 {
-						Errorf("sel: result too large: %d elements", count)
-					}
-					result := make([]Value, 0, count)
-					add := func(howMany, what Value) {
-						hm := int(howMany.(Int))
-						if hm < 0 {
-							hm = -hm
-							what = zero
-						}
-						for ; hm > 0; hm-- {
-							result = append(result, what)
-						}
-					}
-					if len(i) == 1 {
-						for _, y := range j {
-							add(i[0], y)
-						}
-					} else {
-						if len(i) != len(j) {
-							Errorf("sel: unequal lengths %d != %d", len(i), len(j))
-						}
-						for x, y := range j {
-							add(i[x], y)
-						}
-					}
-					return NewVector(result)
+					countV, data := u.(*Vector), v.(*Vector)
+					return data.sel(countV, data.Len())
 				},
 				matrixType: func(c Context, u, v Value) Value {
-					return v.(*Matrix).sel(c, u.(Vector))
+					count, m := u.(*Matrix), v.(*Matrix)
+					if len(count.shape) != 1 {
+						Errorf("sel count cannot be matrix")
+					}
+					result := m.data.sel(count.data, m.shape[len(m.shape)-1])
+					newShape := make([]int, len(m.shape))
+					copy(newShape, m.shape)
+					newShape[len(m.shape)-1] = result.Len() / size(m.shape[:len(m.shape)-1])
+					return NewMatrix(newShape, result)
+				},
+			},
+		},
+
+		{
+			name:      "part",
+			whichType: atLeastVectorType,
+			fn: [numType]binaryFn{
+				vectorType: func(c Context, u, v Value) Value {
+					return v.(*Vector).partition(u.(*Vector))
+				},
+				matrixType: func(c Context, u, v Value) Value {
+					return v.(*Matrix).partition(u.(*Matrix))
 				},
 			},
 		},
@@ -1551,7 +1584,7 @@ func init() {
 			whichType: vectorAndMatrixType,
 			fn: [numType]binaryFn{
 				matrixType: func(c Context, u, v Value) Value {
-					m := v.(*Matrix).binaryTranspose(c, u.(Vector))
+					m := v.(*Matrix).binaryTranspose(c, u.(*Vector))
 					if m.Rank() <= 1 {
 						return m.Data()
 					}
@@ -1561,6 +1594,36 @@ func init() {
 		},
 
 		// Special cases that mix types, so don't promote them.
+		{
+			name:      "===",
+			whichType: noPromoteType,
+			fn: [numType]binaryFn{
+				intType:      equal,
+				charType:     equal,
+				bigIntType:   equal,
+				bigRatType:   equal,
+				bigFloatType: equal,
+				complexType:  equal,
+				vectorType:   equal,
+				matrixType:   equal,
+			},
+		},
+
+		{
+			name:      "!==",
+			whichType: noPromoteType,
+			fn: [numType]binaryFn{
+				intType:      notEqual,
+				charType:     notEqual,
+				bigIntType:   notEqual,
+				bigRatType:   notEqual,
+				bigFloatType: notEqual,
+				complexType:  notEqual,
+				vectorType:   notEqual,
+				matrixType:   notEqual,
+			},
+		},
+
 		{
 			name:      "intersect",
 			whichType: noPromoteType,
