@@ -5,6 +5,8 @@
 package value
 
 import (
+	"fmt"
+	"iter"
 	"math/big"
 	"runtime"
 	"strings"
@@ -38,6 +40,20 @@ type unaryOp struct {
 	fn          [numType]unaryFn
 }
 
+// TraceUnary prints a trace line for a unary operator.
+func TraceUnary(c Context, level int, op string, v Value) {
+	if c.Config().Tracing(level) {
+		fmt.Fprintf(c.Config().ErrOutput(), "\t%s> %s %s\n", c.TraceIndent(), op, v)
+	}
+}
+
+// TraceBinary prints a trace line for a binary operator.
+func TraceBinary(c Context, level int, u Value, op string, v Value) {
+	if c.Config().Tracing(level) {
+		fmt.Fprintf(c.Config().ErrOutput(), "\t%s> %s %s %s\n", c.TraceIndent(), u, op, v)
+	}
+}
+
 func (op *unaryOp) EvalUnary(c Context, v Value) Value {
 	which := whichType(v)
 	fn := op.fn[which]
@@ -52,6 +68,9 @@ func (op *unaryOp) EvalUnary(c Context, v Value) Value {
 		}
 		Errorf("unary %s not implemented on type %s", op.name, which)
 	}
+	if c.Config().Tracing(2) {
+		fmt.Printf("\t%s> %s %s\n", c.TraceIndent(), op.name, v)
+	}
 	return fn(c, v)
 }
 
@@ -78,7 +97,7 @@ func whichType(v Value) valueType {
 		return bigFloatType
 	case Complex:
 		return complexType
-	case Vector:
+	case *Vector:
 		return vectorType
 	case *Matrix:
 		return matrixType
@@ -104,6 +123,9 @@ func (op *binaryOp) EvalBinary(c Context, u, v Value) Value {
 		}
 		Errorf("binary %s not implemented on type %s", op.name, whichV)
 	}
+	if conf.Tracing(2) {
+		fmt.Printf("\t%s> %s %s %s\n", c.TraceIndent(), u, op.name, v)
+	}
 	return fn(c, u, v)
 }
 
@@ -220,49 +242,49 @@ func sendRecover(c chan<- interface{}) {
 // u and v are known to be the same type and at least Vectors.
 func innerProduct(c Context, u Value, left, right string, v Value) Value {
 	switch u := u.(type) {
-	case Vector:
-		v := v.(Vector)
+	case *Vector:
+		v := v.(*Vector)
 		u.sameLength(v)
-		n := len(u)
+		n := u.Len()
 		if n == 0 {
 			Errorf("empty inner product")
 		}
-		x := c.EvalBinary(u[n-1], right, v[n-1])
+		x := c.EvalBinary(u.At(n-1), right, v.At(n-1))
 		for k := n - 2; k >= 0; k-- {
-			x = c.EvalBinary(c.EvalBinary(u[k], right, v[k]), left, x)
+			x = c.EvalBinary(c.EvalBinary(u.At(k), right, v.At(k)), left, x)
 		}
 		return x
 	case *Matrix:
 		// Say we're doing +.*
 		// result[i,j] = +/(u[row i] * v[column j])
 		// Number of columns of u must be the number of rows of v: (-1 take rho u) == (1 take rho v)
-		// The result is has shape (-1 drop rho u), (1 drop rho v)
+		// The result has shape (-1 drop rho u), (1 drop rho v)
 		v := v.(*Matrix)
 		if u.Rank() < 1 || v.Rank() < 1 || u.shape[len(u.shape)-1] != v.shape[0] {
 			Errorf("inner product: mismatched shapes %s and %s", NewIntVector(u.shape...), NewIntVector(v.shape...))
 		}
 		n := v.shape[0]
-		vstride := len(v.data) / n
-		data := make(Vector, len(u.data)/n*vstride)
-		pfor(safeBinary(left) && safeBinary(right), 1, len(data), func(lo, hi int) {
+		vstride := v.data.Len() / n
+		data := newVectorEditor(u.data.Len()/n*vstride, nil)
+		pfor(safeBinary(left) && safeBinary(right), 1, data.Len(), func(lo, hi int) {
 			for x := lo; x < hi; x++ {
 				i := x / vstride * n
 				j := x % vstride
-				acc := c.EvalBinary(u.data[i+n-1], right, v.data[j+(n-1)*vstride])
+				acc := c.EvalBinary(u.data.At(i+n-1), right, v.data.At(j+(n-1)*vstride))
 				for k := n - 2; k >= 0; k-- {
-					acc = c.EvalBinary(c.EvalBinary(u.data[i+k], right, v.data[j+k*vstride]), left, acc)
+					acc = c.EvalBinary(c.EvalBinary(u.data.At(i+k), right, v.data.At(j+k*vstride)), left, acc)
 				}
-				data[x] = acc
+				data.Set(x, acc)
 			}
 		})
 		rank := len(u.shape) + len(v.shape) - 2
 		if rank == 1 {
-			return data
+			return data.Publish()
 		}
 		shape := make([]int, rank)
 		copy(shape, u.shape[:len(u.shape)-1])
 		copy(shape[len(u.shape)-1:], v.shape[1:])
-		return NewMatrix(shape, data)
+		return NewMatrix(shape, data.Publish())
 	}
 	Errorf("can't do inner product on %s", whichType(u))
 	panic("not reached")
@@ -272,32 +294,26 @@ func innerProduct(c Context, u Value, left, right string, v Value) Value {
 // u and v are known to be at least Vectors.
 func outerProduct(c Context, u Value, op string, v Value) Value {
 	switch u := u.(type) {
-	case Vector:
-		v := v.(Vector)
-		m := Matrix{
-			shape: []int{len(u), len(v)},
-			data:  make([]Value, len(u)*len(v)),
-		}
-		pfor(safeBinary(op), 1, len(m.data), func(lo, hi int) {
+	case *Vector:
+		v := v.(*Vector)
+		data := newVectorEditor(u.Len()*v.Len(), nil)
+		pfor(safeBinary(op), 1, data.Len(), func(lo, hi int) {
 			for x := lo; x < hi; x++ {
-				m.data[x] = c.EvalBinary(u[x/len(v)], op, v[x%len(v)])
+				data.Set(x, c.EvalBinary(u.At(x/v.Len()), op, v.At(x%v.Len())))
 			}
 		})
-		return &m // TODO: Shrink?
+		return NewMatrix([]int{u.Len(), v.Len()}, data.Publish())
 	case *Matrix:
 		v := v.(*Matrix)
-		m := Matrix{
-			shape: append(u.Shape(), v.Shape()...),
-			data:  make([]Value, len(u.Data())*len(v.Data())),
-		}
-		vdata := v.Data()
 		udata := u.Data()
-		pfor(safeBinary(op), 1, len(m.data), func(lo, hi int) {
+		vdata := v.Data()
+		data := newVectorEditor(udata.Len()*vdata.Len(), nil)
+		pfor(safeBinary(op), 1, data.Len(), func(lo, hi int) {
 			for x := lo; x < hi; x++ {
-				m.data[x] = c.EvalBinary(udata[x/len(vdata)], op, vdata[x%len(vdata)])
+				data.Set(x, c.EvalBinary(udata.At(x/vdata.Len()), op, vdata.At(x%vdata.Len())))
 			}
 		})
-		return &m // TODO: Shrink?
+		return NewMatrix(append(u.Shape(), v.Shape()...), data.Publish())
 	}
 	Errorf("can't do outer product on %s", whichType(u))
 	panic("not reached")
@@ -310,13 +326,13 @@ func Reduce(c Context, op string, v Value) Value {
 	switch v := v.(type) {
 	case Int, BigInt, BigRat, BigFloat, Complex:
 		return v
-	case Vector:
-		if len(v) == 0 {
+	case *Vector:
+		if v.Len() == 0 {
 			return v
 		}
-		acc := v[len(v)-1]
-		for i := len(v) - 2; i >= 0; i-- {
-			acc = c.EvalBinary(v[i], op, acc)
+		acc := v.At(v.Len() - 1)
+		for i := v.Len() - 2; i >= 0; i-- {
+			acc = c.EvalBinary(v.At(i), op, acc)
 		}
 		return acc
 	case *Matrix:
@@ -328,24 +344,24 @@ func Reduce(c Context, op string, v Value) Value {
 			Errorf("shape for matrix is degenerate: %s", NewIntVector(v.shape...))
 		}
 		shape := v.shape[:v.Rank()-1]
-		data := make(Vector, size(shape))
-		pfor(safeBinary(op), stride, len(data), func(lo, hi int) {
+		data := newVectorEditor(size(shape), nil)
+		pfor(safeBinary(op), stride, data.Len(), func(lo, hi int) {
 			for i := lo; i < hi; i++ {
 				index := stride * i
 				pos := index + stride - 1
-				acc := v.data[pos]
+				acc := v.data.At(pos)
 				pos--
 				for j := 1; j < stride; j++ {
-					acc = c.EvalBinary(v.data[pos], op, acc)
+					acc = c.EvalBinary(v.data.At(pos), op, acc)
 					pos--
 				}
-				data[i] = acc
+				data.Set(i, acc)
 			}
 		})
-		if len(shape) == 1 { // TODO: Matrix.shrink()?
-			return NewVector(data)
+		if len(shape) == 1 {
+			return data.Publish()
 		}
-		return NewMatrix(shape, data)
+		return NewMatrix(shape, data.Publish())
 	}
 	Errorf("can't do reduce on %s", whichType(v))
 	panic("not reached")
@@ -364,26 +380,29 @@ func ReduceFirst(c Context, op string, v Value) Value {
 	if v.Rank() < 2 {
 		Errorf("shape for matrix is degenerate: %s", NewIntVector(m.shape...))
 	}
+	if m.shape[0] == 0 {
+		Errorf("shape for matrix is degenerate: %s", NewIntVector(m.shape...))
+	}
 	stride := size(m.shape[1:m.Rank()])
 	if stride == 0 {
 		Errorf("shape for matrix is degenerate: %s", NewIntVector(m.shape...))
 	}
 	shape := m.shape[1:m.Rank()]
-	data := make(Vector, size(shape))
-	pfor(safeBinary(op), stride, len(data), func(lo, hi int) {
+	data := newVectorEditor(size(shape), nil)
+	pfor(safeBinary(op), stride, data.Len(), func(lo, hi int) {
 		for i := lo; i < hi; i++ {
-			pos := i + len(m.data) - stride
-			acc := m.data[pos]
+			pos := i + m.data.Len() - stride
+			acc := m.data.At(pos)
 			for j := pos - stride; j >= 0; j -= stride {
-				acc = c.EvalBinary(m.data[j], op, acc)
+				acc = c.EvalBinary(m.data.At(j), op, acc)
 			}
-			data[i] = acc
+			data.Set(i, acc)
 		}
 	})
 	if len(shape) == 1 { // TODO: Matrix.shrink()?
-		return NewVector(data)
+		return data.Publish()
 	}
-	return NewMatrix(shape, data)
+	return NewMatrix(shape, data.Publish())
 }
 
 // Scan computes a scan of the op; the \ has been removed.
@@ -393,24 +412,24 @@ func Scan(c Context, op string, v Value) Value {
 	switch v := v.(type) {
 	case Int, BigInt, BigRat, BigFloat, Complex:
 		return v
-	case Vector:
-		if len(v) == 0 {
+	case *Vector:
+		if v.Len() == 0 {
 			return v
 		}
-		values := make(Vector, len(v))
+		values := newVectorEditor(v.Len(), nil)
 		// This is fundamentally O(n²) in the general case.
 		// We make it O(n) for known associative ops.
-		values[0] = v[0]
+		values.Set(0, v.At(0))
 		if knownAssoc(op) {
-			for i := 1; i < len(v); i++ {
-				values[i] = c.EvalBinary(values[i-1], op, v[i])
+			for i := 1; i < v.Len(); i++ {
+				values.Set(i, c.EvalBinary(values.At(i-1), op, v.At(i)))
 			}
 		} else {
-			for i := 1; i < len(v); i++ {
-				values[i] = Reduce(c, op, v[:i+1])
+			for i := 1; i < v.Len(); i++ {
+				values.Set(i, Reduce(c, op, NewVectorSeq(v.Slice(0, i+1))))
 			}
 		}
-		return NewVector(values)
+		return values.Publish()
 	case *Matrix:
 		if v.Rank() < 2 {
 			Errorf("shape for matrix is degenerate: %s", NewIntVector(v.shape...))
@@ -419,26 +438,26 @@ func Scan(c Context, op string, v Value) Value {
 		if stride == 0 {
 			Errorf("shape for matrix is degenerate: %s", NewIntVector(v.shape...))
 		}
-		data := make(Vector, len(v.data))
+		data := newVectorEditor(v.data.Len(), nil)
 		nrows := size(v.shape[:len(v.shape)-1])
 		pfor(safeBinary(op), stride, nrows, func(lo, hi int) {
 			for i := lo; i < hi; i++ {
 				index := i * stride
 				// This is fundamentally O(n²) in the general case.
 				// We make it O(n) for known associative ops.
-				data[index] = v.data[index]
+				data.Set(index, v.data.At(index))
 				if knownAssoc(op) {
 					for j := 1; j < stride; j++ {
-						data[index+j] = c.EvalBinary(data[index+j-1], op, v.data[index+j])
+						data.Set(index+j, c.EvalBinary(data.At(index+j-1), op, v.data.At(index+j)))
 					}
 				} else {
 					for j := 1; j < stride; j++ {
-						data[index+j] = Reduce(c, op, v.data[index:index+j+1])
+						data.Set(index+j, Reduce(c, op, NewVectorSeq(v.data.Slice(index, index+j+1))))
 					}
 				}
 			}
 		})
-		return NewMatrix(v.shape, data)
+		return NewMatrix(v.shape, data.Publish())
 	}
 	Errorf("can't do scan on %s", whichType(v))
 	panic("not reached")
@@ -470,116 +489,274 @@ func ScanFirst(c Context, op string, v Value) Value {
 
 }
 
+// dataShape returns the data shape of v.
+// The data shape of a scalar is []int{}, to distinguish from a vector of length 1.
+func dataShape(v Value) []int {
+	switch v := v.(type) {
+	case *Vector:
+		return []int{v.Len()}
+	case *Matrix:
+		return v.Shape()
+	}
+	return []int{}
+}
+
+// eachOne returns an iterator that yields v once.
+func eachOne(v Value) iter.Seq[Value] {
+	return func(yield func(Value) bool) {
+		yield(v)
+	}
+}
+
+// eachVector returns an iterator that yields each element of v.
+func eachVector(v *Vector) iter.Seq[Value] {
+	return func(yield func(Value) bool) {
+		for _, x := range v.All() {
+			if !yield(x) {
+				break
+			}
+		}
+	}
+}
+
+// eachMatrix returns an iterator that yields subparts of m,
+// iterating over the first dim dimensions of m.
+// If dim == len(m.shape), the iterator yields each value in m.
+// If dim == len(m.shape)-1, the iterator yields each innermost row of m.
+// Otherwise the iterator yields each submatrix obtained by indexing
+// the first dim dimensions of m.
+func eachMatrix(m *Matrix, dim int) iter.Seq[Value] {
+	if dim == len(m.shape) {
+		return eachVector(m.data)
+	}
+	size := m.data.Len()
+	if size > 0 {
+		for d := 0; d < dim; d++ {
+			size /= m.shape[d]
+		}
+	}
+	return func(yield func(Value) bool) {
+		for i := 0; i < m.data.Len(); i += size {
+			var v Value
+			if dim == len(m.shape)-1 {
+				v = NewVectorSeq(m.data.Slice(i, i+size))
+			} else {
+				v = NewMatrix(m.shape[dim:], NewVectorSeq(m.data.Slice(i, i+size)))
+			}
+			if !yield(v) {
+				break
+			}
+		}
+	}
+}
+
+// eachAny returns an iterator that yields subparts of v
+// iterating over the first dim dimensions of v.
+// The caller has checked that dim is in range for v.
+func eachValue(v Value, dim int) iter.Seq[Value] {
+	if dim == 0 {
+		return eachOne(v)
+	}
+	switch v := v.(type) {
+	default:
+		return eachOne(v)
+	case QuietValue:
+		return eachValue(v.Value, dim)
+	case *Vector:
+		if dim != 1 {
+			panic("impossible eachValue")
+		}
+		return eachVector(v)
+	case *Matrix:
+		if dim > len(v.shape) {
+			panic("impossible eachValue")
+		}
+		return eachMatrix(v, dim)
+	}
+}
+
+// BinaryEach computes the result of applying op to lv and rv,
+// applying the "each" expansion depending on how many times
+// @ appears on the left and right ends of op.
+func BinaryEach(c Context, lv Value, op string, rv Value) Value {
+	// Count as many @s as possible for the left side.
+	// Must strip at least one if present, but don't have to strip all,
+	// in case we are doing @ of vector of vectors.
+	l := lv.toType(op, c.Config(), matrixType).(*Matrix)
+	lmax := len(l.shape)
+	ld := 0
+	for ld < len(op) && ld < lmax && op[ld] == '@' {
+		ld++
+	}
+	if ld == 0 && op[0] == '@' {
+		Errorf("%s: left side is scalar", op)
+	}
+	lhs := eachValue(lv, ld)
+
+	// Count as many @s as possible for the right side.
+	// Must strip at least one if present, but don't have to strip all,
+	// in case we are doing @ of vector of vectors.
+	r := rv.toType(op, c.Config(), matrixType).(*Matrix)
+	rmax := len(r.shape)
+	rd := 0
+	for rd < len(op) && rd < rmax && op[len(op)-1-rd] == '@' {
+		rd++
+	}
+	if rd == 0 && op[len(op)-1] == '@' {
+		Errorf("%s: right side is scalar", op)
+	}
+	rhs := eachValue(rv, rd)
+
+	innerOp := op[ld : len(op)-rd]
+	data := newVectorEditor(0, nil)
+
+	for x := range lhs {
+		for y := range rhs {
+			data.Append(c.EvalBinary(x, innerOp, y))
+		}
+	}
+
+	if ld+rd == 1 {
+		return data.Publish()
+	}
+	shape := append(append([]int{}, l.shape[:ld]...), r.shape[:rd]...)
+	return NewMatrix(shape, data.Publish())
+}
+
+// Each computes the result of running op on each element of v.
+// The trailing @ has been removed.
+func Each(c Context, op string, v Value) Value {
+	m := v.toType(op, c.Config(), matrixType).(*Matrix)
+	max := len(m.shape)
+	d := 0
+	for d < len(op) && d < max && op[len(op)-1-d] == '@' {
+		d++
+	}
+	if d == 0 {
+		Errorf("%s: arg is scalar", op)
+	}
+
+	data := newVectorEditor(0, nil)
+	for x := range eachValue(v, d) {
+		data.Append(c.EvalUnary(op[:len(op)-d], x))
+	}
+
+	if d == 1 {
+		return data.Publish()
+	}
+	return NewMatrix(m.shape[:d], data.Publish())
+}
+
 // unaryVectorOp applies op elementwise to i.
 func unaryVectorOp(c Context, op string, i Value) Value {
-	u := i.(Vector)
-	n := make([]Value, len(u))
-	pfor(safeUnary(op), 1, len(n), func(lo, hi int) {
+	u := i.(*Vector)
+	n := newVectorEditor(u.Len(), nil)
+	pfor(safeUnary(op), 1, n.Len(), func(lo, hi int) {
 		for k := lo; k < hi; k++ {
-			n[k] = c.EvalUnary(op, u[k])
+			n.Set(k, c.EvalUnary(op, u.At(k)))
 		}
 	})
-	return NewVector(n)
+	return n.Publish()
 }
 
 // unaryMatrixOp applies op elementwise to i.
 func unaryMatrixOp(c Context, op string, i Value) Value {
 	u := i.(*Matrix)
-	n := make([]Value, len(u.data))
-	pfor(safeUnary(op), 1, len(n), func(lo, hi int) {
+	n := newVectorEditor(u.data.Len(), nil)
+	pfor(safeUnary(op), 1, n.Len(), func(lo, hi int) {
 		for k := lo; k < hi; k++ {
-			n[k] = c.EvalUnary(op, u.data[k])
+			n.Set(k, c.EvalUnary(op, u.data.At(k)))
 		}
 	})
-	return NewMatrix(u.shape, NewVector(n))
+	return NewMatrix(u.shape, n.Publish())
 }
 
 // binaryVectorOp applies op elementwise to i and j.
 func binaryVectorOp(c Context, i Value, op string, j Value) Value {
-	u, v := i.(Vector), j.(Vector)
-	if len(u) == 1 {
-		n := make([]Value, len(v))
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+	u, v := i.(*Vector), j.(*Vector)
+	if u.Len() == 1 {
+		n := newVectorEditor(v.Len(), nil)
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u[0], op, v[k])
+				n.Set(k, c.EvalBinary(u.At(0), op, v.At(k)))
 			}
 		})
-		return NewVector(n)
+		return n.Publish()
 	}
-	if len(v) == 1 {
-		n := make([]Value, len(u))
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+	if v.Len() == 1 {
+		n := newVectorEditor(u.Len(), nil)
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u[k], op, v[0])
+				n.Set(k, c.EvalBinary(u.At(k), op, v.At(0)))
 			}
 		})
-		return NewVector(n)
+		return n.Publish()
 	}
 	u.sameLength(v)
-	n := make([]Value, len(u))
-	pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+	n := newVectorEditor(u.Len(), nil)
+	pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 		for k := lo; k < hi; k++ {
-			n[k] = c.EvalBinary(u[k], op, v[k])
+			n.Set(k, c.EvalBinary(u.At(k), op, v.At(k)))
 		}
 	})
-	return NewVector(n)
+	return n.Publish()
 }
 
 // binaryMatrixOp applies op elementwise to i and j.
 func binaryMatrixOp(c Context, i Value, op string, j Value) Value {
 	u, v := i.(*Matrix), j.(*Matrix)
 	shape := u.shape
-	var n []Value
+	var n *vectorEditor
 
 	// One or the other may be a scalar in disguise.
 	switch {
 	case isScalar(u):
 		// Scalar op Matrix.
 		shape = v.shape
-		n = make([]Value, len(v.data))
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+		n = newVectorEditor(v.data.Len(), nil)
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u.data[0], op, v.data[k])
+				n.Set(k, c.EvalBinary(u.data.At(0), op, v.data.At(k)))
 			}
 		})
 	case isScalar(v):
 		// Matrix op Scalar.
-		n = make([]Value, len(u.data))
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+		n = newVectorEditor(u.data.Len(), nil)
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u.data[k], op, v.data[0])
+				n.Set(k, c.EvalBinary(u.data.At(k), op, v.data.At(0)))
 			}
 		})
 	case isVector(u, v.shape):
 		// Vector op Matrix.
 		shape = v.shape
-		n = make([]Value, len(v.data))
+		n = newVectorEditor(v.data.Len(), nil)
 		dim := u.shape[0]
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u.data[k%dim], op, v.data[k])
+				n.Set(k, c.EvalBinary(u.data.At(k%dim), op, v.data.At(k)))
 			}
 		})
 	case isVector(v, u.shape):
 		// Matrix op Vector.
-		n = make([]Value, len(u.data))
+		n = newVectorEditor(u.data.Len(), nil)
 		dim := v.shape[0]
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u.data[k], op, v.data[k%dim])
+				n.Set(k, c.EvalBinary(u.data.At(k), op, v.data.At(k%dim)))
 			}
 		})
 	default:
 		// Matrix op Matrix.
 		u.sameShape(v)
-		n = make([]Value, len(u.data))
-		pfor(safeBinary(op), 1, len(n), func(lo, hi int) {
+		n = newVectorEditor(u.data.Len(), nil)
+		pfor(safeBinary(op), 1, n.Len(), func(lo, hi int) {
 			for k := lo; k < hi; k++ {
-				n[k] = c.EvalBinary(u.data[k], op, v.data[k])
+				n.Set(k, c.EvalBinary(u.data.At(k), op, v.data.At(k)))
 			}
 		})
 	}
-	return NewMatrix(shape, NewVector(n))
+	return NewMatrix(shape, n.Publish())
 }
 
 // IsScalarType reports whether u is an actual scalar, an int or float etc.
@@ -674,7 +851,7 @@ func compare(v Value, i int) int {
 }
 
 // isTrue reports whether v represents boolean truth. If v is not
-// a scalar, an error results.
+// ultimately a scalar, an error results.
 func isTrue(fnName string, v Value) bool {
 	switch i := v.(type) {
 	case Char:
@@ -689,10 +866,17 @@ func isTrue(fnName string, v Value) bool {
 		return i.Float.Sign() != 0
 	case Complex:
 		return !isZero(v)
-	default:
-		Errorf("invalid expression %s for conditional inside %q", v, fnName)
-		return false
+	case *Vector:
+		if i.Len() == 1 {
+			return isTrue(fnName, i.At(0))
+		}
+	case *Matrix:
+		if i.data.Len() == 1 {
+			return isTrue(fnName, i.data.At(0))
+		}
 	}
+	Errorf("invalid expression %s for conditional inside %q", v, fnName)
+	return false
 }
 
 // sgn is a wrapper for calling "sgn v".
@@ -726,8 +910,10 @@ func mod(c Context, a, b Value) Value {
 // QuoRem uses Euclidean division to return the quotient and remainder for a/b.
 // The quotient will be an integer, possibly negative; the remainder is always positive
 // and may be fractional. Returned values satisfy the identity that
+//
 //	quo = a div b  such that
 //	rem = a - b*quo  with 0 <= rem < |y|
+//
 // See comment for math/big.Int.DivMod for details.
 // Exported for testing.
 func QuoRem(op string, c Context, a, b Value) (div, rem Value) {
@@ -846,31 +1032,38 @@ func EvalFunctionBody(context Context, fnName string, body []Expr) Value {
 }
 
 // flatten returns a simple vector containing the scalar elements of v.
-func flatten(v Value) []Value {
-	if IsScalarType(v) {
-		return []Value{v}
+func flatten(v Value) iter.Seq2[int, Value] {
+	return func(yield func(int, Value) bool) {
+		flattenTo(0, v, yield)
 	}
+}
+
+// flattenTo flattens the values contained in v,
+// calling yield(off, x0), yield(off+1, x1), ... for successive values.
+// It returns the new next offset to use
+// and whether the iteration should continue at all.
+func flattenTo(off int, v Value, yield func(int, Value) bool) (newOff int, cont bool) {
 	switch v := v.(type) {
-	case Vector:
-		if v.allScalars() {
-			return v.Copy().(Vector)
-		}
-		elems := make([]Value, 0, len(v))
-		for _, elem := range v {
-			if IsScalarType(elem) {
-				elems = append(elems, elem)
-			} else {
-				elems = append(elems, flatten(elem)...)
-			}
-		}
-		return elems
 	case *Matrix:
-		return flatten(v.data)
+		return flattenTo(off, v.data, yield)
+	case *Vector:
+		for _, elem := range v.All() {
+			newOff, cont := flattenTo(off, elem, yield)
+			if !cont {
+				return 0, false
+			}
+			off = newOff
+		}
+		return off, true
+	default:
+		if !yield(off, v) {
+			return 0, false
+		}
+		return off + 1, true
 	}
-	panic("flatten: can't happen")
 }
 
 // box returns its argument wrapped into a one-element vector.
 func box(c Context, v Value) Value {
-	return NewVector([]Value{v})
+	return NewVector(v)
 }
