@@ -4,17 +4,23 @@
 
 package value
 
-import "sync"
+import (
+	"slices"
+	"sync"
+)
 
 // An indexState holds the state needed to locate
 // the values denoted by an index expression left[index],
 // which is evaluated to lhs[indexes].
 type indexState struct {
-	lhs   Value
-	slice []Value // underlying data slice for lhs
-	shape []int   // underlying shape for lhs
+	lhs    Value
+	vector *Vector
+	edit   *vectorEditor
+	shape  []int // underlying shape for lhs
 
-	indexes []Vector // Vectors of all Int, all in range for shape
+	indexes     []*Vector // Vectors of all Int, all in range for shape...
+	indexVector *Vector   // ... or vector of index vectors
+	indexDim    int       // number of dimensions indexed
 
 	outShape []int // output shape (nil is scalar)
 	outSize  int   // output size (# scalars)
@@ -23,18 +29,21 @@ type indexState struct {
 // init initializes ix to describe top, which is left[index].
 // Left and index will be evaluated (right to left),
 // while top is only for its ProgString method.
-func (ix *indexState) init(context Context, top, left Expr, index []Expr) {
+// If lvarx is not nil, then it is the variable expression
+// corresponding to left, to be used for assignments.
+func (ix *indexState) init(context Context, top, left Expr, lvarx *VarExpr, index []Expr) {
 	// Evaluate indexes, make sure all are Vector of Int.
 	// Compute shape of result as we go.
 	// Scalar indexes drop a dimension,
 	// while vector and matrix indexes replace the dimension with their shape.
-	ix.indexes = make([]Vector, len(index))
+	ix.indexes = make([]*Vector, len(index))
 	ix.outShape = nil          // common case - scalar indexes covering entire rank → scalar result
 	var outShapeToUpdate []int // indexes of outShape entries that need updating after lhs eval.
+	missing := make([]bool, len(index))
 	for i := len(index) - 1; i >= 0; i-- {
 		if index[i] == nil {
 			// Make this iota(dimension), to be filled in after evaluating lhs.
-			ix.indexes[i] = nil
+			missing[i] = true
 			outShapeToUpdate = append(outShapeToUpdate, len(ix.outShape))
 			ix.outShape = append(ix.outShape, 0) // Fixed below, after we have evaluated left.
 			continue
@@ -42,12 +51,12 @@ func (ix *indexState) init(context Context, top, left Expr, index []Expr) {
 		x := index[i].Eval(context).Inner()
 		switch x := x.(type) {
 		default:
-			Errorf("invalid index %s (%s) in %s", index[i].ProgString(), whichType(x), top.ProgString())
+			Errorf("invalid index %s (type %s) in %s", index[i].ProgString(), whichType(x), top.ProgString())
 		case Int:
-			ix.indexes[i] = Vector{x}
-		case Vector:
+			ix.indexes[i] = NewVector(x)
+		case *Vector:
 			ix.indexes[i] = x
-			ix.outShape = append(ix.outShape, len(x))
+			ix.outShape = append(ix.outShape, x.Len())
 		case *Matrix:
 			ix.indexes[i] = x.Data()
 			// Append shape in reverse, because ix.shape will be reversed below.
@@ -56,17 +65,11 @@ func (ix *indexState) init(context Context, top, left Expr, index []Expr) {
 				ix.outShape = append(ix.outShape, shape[j])
 			}
 		}
-		for _, v := range ix.indexes[i] {
-			if _, ok := v.(Int); !ok {
-				Errorf("invalid index %s (%s) in %s in %s", v, whichType(v), index[i].ProgString(), top.ProgString())
-			}
-		}
 	}
 
 	// Walked indexes right-to-left, so reverse shape.
-	for i, j := 0, len(ix.outShape)-1; i < j; i, j = i+1, j-1 {
-		ix.outShape[i], ix.outShape[j] = ix.outShape[j], ix.outShape[i]
-	}
+	slices.Reverse(ix.outShape)
+
 	// The offsets stored in outShapeToUpdate must also be flipped.
 	for i, o := range outShapeToUpdate {
 		outShapeToUpdate[i] = len(ix.outShape) - o - 1
@@ -74,40 +77,69 @@ func (ix *indexState) init(context Context, top, left Expr, index []Expr) {
 
 	// Can now safely evaluate left side
 	// (must wait until indexes have been evaluated, R-to-L).
-	ix.lhs = left.Eval(context)
+	var lvar *Var
+	if lvarx != nil {
+		if lvarx.Local >= 1 {
+			lvar = context.Local(lvarx.Local)
+		} else {
+			lvar = context.Global(lvarx.Name)
+			if lvar == nil {
+				Errorf("undefined global variable %q", lvarx.Name)
+			}
+		}
+		ix.lhs = lvar.value
+	} else {
+		ix.lhs = left.Eval(context)
+	}
 	switch lhs := ix.lhs.(type) {
 	default:
 		Errorf("cannot index %s (%v)", left.ProgString(), whichType(lhs))
 	case *Matrix:
-		ix.slice = lhs.Data()
+		ix.vector = lhs.data
+		if lvar != nil {
+			ix.edit = lvar.editor()
+		}
 		ix.shape = lhs.Shape()
-	case Vector:
-		ix.slice = lhs
-		ix.shape = []int{len(lhs)}
+	case *Vector:
+		ix.vector = lhs
+		if lvar != nil {
+			ix.edit = lvar.editor()
+		}
+		ix.shape = []int{lhs.Len()}
+	}
+
+	origin := Int(context.Config().Origin())
+	if ix.initVectorIndex(origin) {
+		return
 	}
 
 	// Finish the result shape.
-	origin := Int(context.Config().Origin())
 	if len(ix.indexes) > len(ix.shape) {
 		Errorf("too many dimensions in %s indexing shape %v", top.ProgString(), NewIntVector(ix.shape...))
 	}
 	// Replace nil index entries, created above, with iota(dimension).
 	j := 0
-	for i, v := range ix.indexes {
-		if v == nil {
-			x := constIota(int(origin), ix.shape[i])
-			ix.indexes[i] = NewVector(x)
-			ix.outShape[outShapeToUpdate[j]] = len(x)
+	for i := range ix.indexes {
+		if missing[i] {
+			x := newIota(int(origin), ix.shape[i])
+			ix.indexes[i] = x
+			ix.outShape[outShapeToUpdate[j]] = x.Len()
 			j++
 		}
 	}
 	ix.outShape = append(ix.outShape, ix.shape[len(index):]...)
 	ix.outSize = size(ix.outShape)
+	ix.indexDim = len(index)
 
 	// Check indexes are all valid.
 	for i, v := range ix.indexes {
-		for j := range v {
-			vj := v[j].(Int)
+		for _, vj := range v.All() {
+			if _, ok := vj.(Int); !ok {
+				Errorf("invalid index %s (type %s) in %s", vj, whichType(vj), top.ProgString())
+			}
+		}
+		for j := range v.All() {
+			vj := v.At(j).(Int)
 			if vj < origin || vj-origin >= Int(ix.shape[i]) {
 				s := left.ProgString() + "["
 				for k := range ix.indexes {
@@ -127,12 +159,54 @@ func (ix *indexState) init(context Context, top, left Expr, index []Expr) {
 	}
 }
 
+func (ix *indexState) initVectorIndex(origin Int) bool {
+	// Check for single index, vector of vectors,
+	// all of which have the same length <= rank of lhs.
+	if len(ix.indexes) != 1 || ix.indexes[0] == nil || ix.indexes[0].Len() == 0 {
+		return false
+	}
+	data := ix.indexes[0]
+	v, ok := data.At(0).(*Vector)
+	if !ok {
+		return false
+	}
+	if v.Len() > len(ix.shape) {
+		Errorf("index vector (%v) too long for shape %v", v, NewIntVector(ix.shape...))
+	}
+	n := v.Len()
+	for _, x := range data.All() {
+		v, ok := x.(*Vector)
+		if !ok {
+			Errorf("mixed vector and non-vector indices %v and %v", data.At(0), x)
+		}
+		if v.Len() != n {
+			Errorf("index vectors of mixed lengths %v and %v", data.At(0), v)
+		}
+		// Check vector content against shape.
+		for j, vj := range v.All() {
+			k, ok := vj.(Int)
+			if !ok {
+				Errorf("index vector %v contains invalid index %v (type %s)", v, vj, whichType(vj))
+			}
+			if k < origin || k-origin >= Int(ix.shape[j]) {
+				Errorf("index vector %v out of range for shape %v", v, NewIntVector(ix.shape...))
+			}
+		}
+	}
+	ix.indexes = nil
+	ix.indexVector = data
+	ix.indexDim = n
+	ix.outShape = append(ix.outShape, ix.shape[n:]...)
+	ix.outSize = size(ix.outShape)
+	return true
+}
+
 // Index returns left[index].
 // Left and index will be evaluated (right to left),
 // while top is only for its ProgString method.
 func Index(context Context, top, left Expr, index []Expr) Value {
 	var ix indexState
-	ix.init(context, top, left, index)
+	ix.init(context, top, left, nil, index)
 	origin := Int(context.Config().Origin())
 
 	if len(ix.outShape) == 0 {
@@ -142,84 +216,103 @@ func Index(context Context, top, left Expr, index []Expr) Value {
 			if j > 0 {
 				offset *= ix.shape[j]
 			}
-			offset += int(ix.indexes[j][0].(Int) - origin)
+			offset += int(ix.indexes[j].At(0).(Int) - origin)
 		}
-		return ix.slice[offset]
+		return ix.vector.At(offset)
 	}
 
-	data := make(Vector, ix.outSize)
-	copySize := int(size(ix.shape[len(ix.indexes):]))
-	n := len(data) / copySize
-	coord := make([]int, len(ix.indexes))
-	for i := 0; i < n; i++ {
-		// Copy data for indexes[coord].
-		offset := 0
-		for j := 0; j < len(ix.indexes); j++ {
-			if j > 0 {
-				offset *= ix.shape[j]
-			}
-			offset += int(ix.indexes[j][coord[j]].(Int) - origin)
-		}
-		copy(data[i*copySize:(i+1)*copySize], ix.slice[offset*copySize:(offset+1)*copySize])
+	data := newVectorEditor(ix.outSize, nil)
+	copySize := int(size(ix.shape[ix.indexDim:]))
+	n := data.Len() / copySize
 
-		// Increment coord.
-		for j := len(coord) - 1; j >= 0; j-- {
-			if coord[j]++; coord[j] < len(ix.indexes[j]) {
-				break
+	if ix.indexVector != nil {
+		for i, v := range ix.indexVector.All() {
+			offset := 0
+			for j, vj := range v.(*Vector).All() {
+				if j > 0 {
+					offset *= ix.shape[j]
+				}
+				offset += int(vj.(Int) - origin)
+			}
+			for k := range copySize {
+				data.Set(i*copySize+k, ix.vector.At(offset*copySize+k))
+			}
+		}
+	} else {
+		coord := make([]int, ix.indexDim)
+		for i := 0; i < n; i++ {
+			// Copy data for indexes[coord].
+			offset := 0
+			for j := 0; j < len(ix.indexes); j++ {
+				if j > 0 {
+					offset *= ix.shape[j]
+				}
+				offset += int(ix.indexes[j].At(coord[j]).(Int) - origin)
+			}
+			for k := range copySize {
+				data.Set(i*copySize+k, ix.vector.At(offset*copySize+k))
+			}
+
+			// Increment coord.
+			for j := len(coord) - 1; j >= 0; j-- {
+				if coord[j]++; coord[j] < ix.indexes[j].Len() {
+					break
+				}
+				coord[j] = 0
 			}
-			coord[j] = 0
 		}
 	}
 
 	if len(ix.outShape) == 0 {
-		return data[0]
+		return data.At(0)
 	}
 	if len(ix.outShape) == 1 {
-		return data
+		return data.Publish()
 	}
-	return NewMatrix(ix.outShape, data)
+	return NewMatrix(ix.outShape, data.Publish())
 }
 
 // IndexAssign handles general assignment to indexed expressions on the LHS.
 // Left and index will be evaluated (right to left),
 // while top is only for its ProgString method.
-// The caller must check that left is a variable expression,
-// so that the assignment is not being written into a temporary.
-func IndexAssign(context Context, top, left Expr, index []Expr, right Expr, rhs Value) {
+// The caller must check that left is a variable expression
+// and pass lvar, the variable corresponding to left.
+func IndexAssign(context Context, top, left Expr, lvarx *VarExpr, index []Expr, right Expr, rhs Value) {
 	var ix indexState
-	ix.init(context, top, left, index)
+	ix.init(context, top, left, lvarx, index)
 
 	// Unless assigning to a single cell, RHS must be scalar or
 	// have same shape as indexed expression.
 	var rscalar Value
-	var rslice []Value
+	var rvector *Vector
 	if len(ix.outShape) == 0 {
 		rscalar = rhs
 	} else {
+		badShape := func(rshape ...int) {
+			var where string
+			if right == nil {
+				where = "to " + top.ProgString()
+			} else {
+				where = top.ProgString() + " = " + right.ProgString()
+			}
+			Errorf("shape mismatch %v != %v in assignment %v",
+				NewIntVector(ix.outShape...), NewIntVector(rshape...),
+				where)
+		}
+
 		switch rhs := rhs.(type) {
 		default:
 			rscalar = rhs
-		case Vector:
-			if len(ix.outShape) != 1 || ix.outShape[0] != len(rhs) {
-				Errorf("shape mismatch %v != %v in assignment %v = %v",
-					NewIntVector(ix.outShape...), NewIntVector(len(rhs)),
-					top.ProgString(), right.ProgString())
+		case *Vector:
+			if len(ix.outShape) != 1 || ix.outShape[0] != rhs.Len() {
+				badShape(rhs.Len())
 			}
-			rslice = rhs
+			rvector = rhs
 		case *Matrix:
 			if !sameShape(ix.outShape, rhs.Shape()) {
-				Errorf("shape mismatch %v != %v in assignment %v = %v",
-					NewIntVector(ix.outShape...), NewIntVector(rhs.Shape()...),
-					top.ProgString(), right.ProgString())
-			}
-			rslice = rhs.Data()
-			if rhs == ix.lhs {
-				// Assigning entire rhs to some permutation of lhs.
-				// Make copy of slice to avoid problems with overwriting
-				// values we need to read later. Uncommon.
-				rslice = make([]Value, len(rslice))
-				copy(rslice, rhs.Data())
+				badShape(rhs.Shape()...)
 			}
+			rvector = rhs.data
 		}
 	}
 
@@ -231,20 +324,44 @@ func IndexAssign(context Context, top, left Expr, index []Expr, right Expr, rhs
 			if j > 0 {
 				offset *= ix.shape[j]
 			}
-			offset += int(ix.indexes[j][0].(Int) - origin)
+			offset += int(ix.indexes[j].At(0).(Int) - origin)
 		}
-		ix.slice[offset] = rscalar.Copy()
+		ix.edit.Set(offset, rscalar)
 		return
 	}
 
-	copySize := int(size(ix.shape[len(ix.indexes):]))
+	copySize := int(size(ix.shape[ix.indexDim:]))
 	n := ix.outSize / copySize
 	pfor(true, copySize, n, func(lo, hi int) {
+		if ix.indexVector != nil {
+			for i := lo; i < hi; i++ {
+				v := ix.indexVector.At(i).(*Vector)
+				offset := 0
+				for j, vj := range v.All() {
+					if j > 0 {
+						offset *= ix.shape[j]
+					}
+					offset += int(vj.(Int) - origin)
+				}
+				dstOff := offset * copySize
+				if rscalar != nil {
+					for j := range copySize {
+						ix.edit.Set(dstOff+j, rscalar)
+					}
+				} else {
+					for j := range copySize {
+						ix.edit.Set(dstOff+j, rvector.At(j+i*copySize))
+					}
+				}
+			}
+			return
+		}
+
 		// Compute starting coordinate index.
 		coord := make([]int, len(ix.indexes))
 		i := lo
 		for j := len(coord) - 1; j >= 0; j-- {
-			if n := len(ix.indexes[j]); n > 0 {
+			if n := ix.indexes[j].Len(); n > 0 {
 				coord[j] = i % n
 				i /= n
 			}
@@ -257,22 +374,22 @@ func IndexAssign(context Context, top, left Expr, index []Expr, right Expr, rhs
 				if j > 0 {
 					offset *= ix.shape[j]
 				}
-				offset += int(ix.indexes[j][coord[j]].(Int) - origin)
+				offset += int(ix.indexes[j].At(coord[j]).(Int) - origin)
 			}
-			dst := ix.slice[offset*copySize : (offset+1)*copySize]
+			dstOff := offset * copySize
 			if rscalar != nil {
-				for j := range dst {
-					dst[j] = rscalar.Copy()
+				for j := range copySize {
+					ix.edit.Set(dstOff+j, rscalar)
 				}
 			} else {
-				for j := range dst {
-					dst[j] = rslice[j+i*copySize].Copy()
+				for j := range copySize {
+					ix.edit.Set(dstOff+j, rvector.At(j+i*copySize))
 				}
 			}
 
 			// Increment coord.
 			for j := len(coord) - 1; j >= 0; j-- {
-				if coord[j]++; coord[j] < len(ix.indexes[j]) {
+				if coord[j]++; coord[j] < ix.indexes[j].Len() {
 					break
 				}
 				coord[j] = 0
@@ -299,7 +416,14 @@ func constIota(origin, n int) []Value {
 		iotaLock.RUnlock()
 		growIota(origin + n)
 	}
+}
 
+// newIota returns the result of 'iota n' as a new Vector.
+func newIota(origin, n int) *Vector {
+	if n < 0 || maxInt < n {
+		Errorf("bad iota %d", n)
+	}
+	return NewVector(constIota(origin, int(n))...)
 }
 
 func growIota(n int) {
