@@ -5,10 +5,10 @@
 package value
 
 import (
-	"bytes"
 	"fmt"
 	"io"
 	"math/bits"
+	"slices"
 	"sort"
 	"strings"
 
@@ -25,7 +25,7 @@ import (
 
 type Matrix struct {
 	shape []int
-	data  Vector
+	data  *Vector
 }
 
 // Shape returns the shape of the matrix.
@@ -39,106 +39,73 @@ func (m *Matrix) Rank() int {
 
 func (m *Matrix) shrink() Value {
 	if len(m.shape) == 1 {
-		return NewVector(m.data)
+		return m.data
 	}
 	return m
 }
 
 // Data returns the data of the matrix as a vector.
-func (m *Matrix) Data() Vector {
+func (m *Matrix) Data() *Vector {
 	return m.data
 }
 
-func (m *Matrix) Copy() Value {
-	shape := make([]int, len(m.shape))
-	data := make([]Value, len(m.data))
-	copy(shape, m.shape)
-	copy(data, m.data)
-	return &Matrix{
-		shape: shape,
-		data:  data,
-	}
+// widths is a type that records the maximum widths of each column,
+// then reports them back as needed. It handles two cases: If all
+// widths are small, it always replies with the maximum width seen;
+// otherwise it replies with the maximum for this column alone.
+type widths struct {
+	wid []int // Maximum width seen in each column.
+	max int   // The maximum over all columns.
 }
 
-// elemStrs returns the formatted elements of the matrix and the width of the widest element.
-// Each element is represented by a slice of lines, that is, the return value is indexed by
-// [elem][line].
-func (m *Matrix) elemStrs(conf *config.Config) ([][]string, int) {
-	// Format the matrix as a vector, and then in write2d we rearrange the pieces.
-	// In the formatting, there's no need for spacing the elements as we'll cut
-	// them apart ourselves using column information. Spaces will be added
-	// when needed in write2d.
-	v := NewVector(m.data)
-	lines, cols := v.multiLineSprint(conf, v.allScalars(), v.AllChars(), !withSpaces, !trimTrailingSpace)
-	strs := make([][]string, len(m.data))
-	wid := 0
-	for i := range m.data {
-		rows := make([]string, len(lines))
-		for j, line := range lines {
-			if i == 0 {
-				rows[j] = line[:cols[0]]
-			} else {
-				rows[j] = line[cols[i-1]:cols[i]]
-			}
-		}
-		if len(rows[0]) > wid {
-			wid = len(rows[0])
-		}
-		strs[i] = rows
+const widthThreshold = 5 // All widths below this -> regular grid.
+
+// addColumn records the width for column i, either updating that
+// column or adding a new one. i is never more than len(w.wid).
+func (w *widths) addColumn(i, wid int) {
+	switch {
+	case i < len(w.wid):
+		w.wid[i] = max(wid, w.wid[i])
+	case i == len(w.wid):
+		w.wid = append(w.wid, wid)
+	default:
+		Errorf("cannot happen: out of range in addColumn")
 	}
-	return strs, wid
+	w.max = max(w.max, wid)
 }
 
-// write2d prints the 2d matrix m into the buffer.
-// elems is a slice (of slices) of already-printed values.
-// The receiver provides only the shape of the matrix.
-func (m *Matrix) write2d(b *bytes.Buffer, elems [][]string, width int) {
-	nrows := m.shape[0]
-	ncols := m.shape[1]
-	index := 0
-	for row := 0; row < nrows; row++ {
-		if row > 0 {
-			b.WriteByte('\n')
-		}
-		// Don't print the line if it has no content.
-		nonBlankLine := 0
-		for col := 0; col < ncols; col++ {
-			strs := elems[index+col]
-			for line := nonBlankLine; line < len(strs); line++ {
-				for _, r := range strs[line] {
-					if r != ' ' {
-						nonBlankLine = line
-						break
-					}
-				}
-			}
-		}
-		for line := 0; line < nonBlankLine+1; line++ {
-			if line > 0 {
-				b.WriteByte('\n')
-			}
-			for col := 0; col < ncols; col++ {
-				str := elems[index+col][line]
-				b.WriteString(blanks(width - len(str)))
-				b.WriteString(str)
-				if (col+1)%ncols != 0 {
-					b.WriteString(" ")
-				}
-			}
-		}
-		index += ncols
+// column returns the width to use to display column i.
+func (w *widths) column(i int) int {
+	if w.max < widthThreshold {
+		return w.max
 	}
+	return w.wid[i]
+}
+
+// columns returns the number of columns.
+func (w *widths) columns() int {
+	return len(w.wid)
+}
+
+// write2d formats the 2d grid elements into lines.
+func (m *Matrix) write2d(elems [][]string, cols int, w *widths) []string {
+	var lines []string
+	for row := range len(elems) / cols {
+		cells := elems[row*cols : (row+1)*cols]
+		lines = append(lines, formatRow(cells, w)...)
+	}
+	return lines
 }
 
 func (m *Matrix) fprintf(c Context, w io.Writer, format string) {
 	rank := len(m.shape)
-	if rank == 0 || len(m.data) == 0 {
+	if rank == 0 || m.data.Len() == 0 {
 		return
 	}
 	counters := make([]int, len(m.shape))
 	verb := verbOf(format)
 	printSpace := false
-	for i, v := range m.data {
+	for i, v := range m.data.All() {
 		if printSpace {
 			fmt.Fprint(w, " ")
 		}
@@ -153,7 +120,7 @@ func (m *Matrix) fprintf(c Context, w io.Writer, format string) {
 			// Each time a counter overflows, add a newline.
 			// This puts 0 lines between rows, 1 between
 			// each 2-d block, 2 between each 3-d block, etc.
-			if i < len(m.data)-1 {
+			if i < m.data.Len()-1 {
 				w.Write([]byte{'\n'})
 				printSpace = false
 			}
@@ -167,66 +134,73 @@ func (m *Matrix) String() string {
 }
 
 func (m *Matrix) Sprint(conf *config.Config) string {
-	var b bytes.Buffer
+	return strings.Join(m.sprint(conf), "\n")
+}
+
+func (m *Matrix) sprint(conf *config.Config) []string {
 	switch m.Rank() {
 	case 0:
-		Errorf("matrix is scalar")
+		Errorf("matrix is rank 0") // Can this ever happen?
+		return nil
 	case 1:
-		Errorf("matrix is vector")
+		return m.data.sprint(conf)
 	case 2:
-		nrows := m.shape[0]
 		ncols := m.shape[1]
+		nrows := m.shape[0]
 		if nrows == 0 || ncols == 0 {
-			return ""
+			return nil
 		}
 		// If it's all chars, print it without padding or quotes.
 		if m.data.AllChars() {
+			var lines []string
 			for i := 0; i < nrows; i++ {
-				if i > 0 {
-					b.WriteByte('\n')
-				}
-				fmt.Fprintf(&b, "%s", m.data[i*ncols:(i+1)*ncols].Sprint(conf))
+				// TODO what about embedded newlines?
+				lines = append(lines, NewVectorSeq(m.data.Slice(i*ncols, (i+1)*ncols)).Sprint(conf))
 			}
-			break
+			return lines
 		}
-		strs, wid := m.elemStrs(conf)
-		m.write2d(&b, strs, wid)
+		cells, width := m.data.cells(conf, ncols)
+		return m.write2d(cells, ncols, width)
 	case 3:
-		// If it's all chars, print it without padding or quotes.
-		if m.data.AllChars() {
-			nelems := m.shape[0]
-			ElemSize := m.ElemSize()
-			index := 0
-			for i := 0; i < nelems; i++ {
-				if i > 0 {
-					b.WriteString("\n\n")
-				}
-				fmt.Fprintf(&b, "%s", NewMatrix(m.shape[1:], m.data[index:index+ElemSize]).Sprint(conf))
-				index += ElemSize
-			}
-			break
-		}
 		// As for 2d: print the vector elements, compute the
 		// global width, and use that to print each 2d submatrix.
-		n2d := m.shape[0]    // number of 2d submatrices.
+		n2d := m.shape[0] // number of 2d submatrices.
+		nrows := m.shape[1]
+		ncols := m.shape[2]
 		size := m.ElemSize() // number of elems in each submatrix.
-		strs, wid := m.elemStrs(conf)
-		start := 0
-		for i := 0; i < n2d; i++ {
+
+		// If it's all chars, print it without padding or quotes.
+		if m.data.AllChars() {
+			var lines []string
+			start := 0
+			for i := range n2d {
+				if i > 0 {
+					lines = append(lines, "")
+				}
+				for range nrows {
+					lines = append(lines, NewVectorSeq(m.data.Slice(start, start+ncols)).sprint(conf)...)
+					start += ncols
+				}
+			}
+			return lines
+		}
+
+		cells, width := m.data.cells(conf, ncols)
+		var lines []string
+		for i := range n2d {
 			if i > 0 {
-				b.WriteString("\n\n")
+				lines = append(lines, "")
 			}
 			m := Matrix{
 				shape: m.shape[1:],
-				data:  m.data[start : start+size],
+				// no data; write2d uses cells, not data
 			}
-			m.write2d(&b, strs[start:start+size], wid)
-			start += size
+			lines = append(lines, m.write2d(cells[i*size:(i+1)*size], ncols, width)...)
 		}
+		return lines
 	default:
 		return m.higherDim(conf, "[", 0)
 	}
-	return b.String()
 }
 
 func (m *Matrix) ProgString() string {
@@ -234,46 +208,36 @@ func (m *Matrix) ProgString() string {
 	panic("matrix.ProgString - cannot happen")
 }
 
-func (m *Matrix) higherDim(conf *config.Config, prefix string, indentation int) string {
+func (m *Matrix) higherDim(conf *config.Config, prefix string, indentation int) []string {
 	if m.Rank() <= 3 {
-		return indent(indentation, m.Sprint(conf))
+		return indent(indentation, m.sprint(conf))
 	}
 	dim := m.shape[0]
 	rest := strings.Repeat(" *", m.Rank()-1)[1:]
-	var b bytes.Buffer
+	var lines []string
 	for i := 0; i < dim; i++ {
 		inner := Matrix{
 			shape: m.shape[1:],
-			data:  m.data[i*m.ElemSize():],
+			data:  NewVectorSeq(m.data.Slice(i*m.ElemSize(), m.data.Len())),
 		}
 		if i > 0 {
-			b.WriteString("\n\n")
+			lines = append(lines, "")
 		}
 		innerPrefix := fmt.Sprintf("%s%d ", prefix, i+conf.Origin())
-		b.WriteString(indent(indentation, "%s%s]:\n", innerPrefix, rest))
-		b.WriteString(inner.higherDim(conf, innerPrefix, indentation+1))
+		lines = append(lines, fmt.Sprintf("%s%s]:", innerPrefix, rest))
+		lines = append(lines, inner.higherDim(conf, innerPrefix, indentation+1)...)
 	}
-	return b.String()
+	return lines
 }
 
-// indent prints the args, indenting each line by the specified amount.
-func indent(indentation int, format string, args ...interface{}) string {
-	s := fmt.Sprintf(format, args...)
-	if indentation == 0 {
-		return s
-	}
-	var b bytes.Buffer
-	lines := strings.SplitAfter(s, "\n")
-	if len(lines) > 0 && lines[len(lines)-1] == "" {
-		lines = lines[:len(lines)-1]
-	}
+// indent add indentation to each element in lines,
+// returning a new slice of lines.
+func indent(indentation int, lines []string) []string {
+	var out []string
 	for _, line := range lines {
-		if len(line) > 0 {
-			b.WriteString(spaces(indentation))
-		}
-		b.WriteString(line)
+		out = append(out, spaces(indentation)+line)
 	}
-	return b.String()
+	return out
 }
 
 // spaces returns 2*n space characters, maxing out at 2*10.
@@ -309,7 +273,7 @@ func size(shape []int) int {
 }
 
 // NewMatrix makes a new matrix. The number of elements must fit in an Int.
-func NewMatrix(shape []int, data []Value) *Matrix {
+func NewMatrix(shape []int, data *Vector) *Matrix {
 	// Check consistency and sanity.
 	nelems := 0
 	if len(shape) > 0 {
@@ -323,8 +287,8 @@ func NewMatrix(shape []int, data []Value) *Matrix {
 		}
 		nelems = int(n)
 	}
-	if nelems != len(data) {
-		Errorf("inconsistent shape (%d) and data size (%d) for new matrix", shape, len(data))
+	if nelems != data.Len() {
+		Errorf("inconsistent shape (%d) and data size (%d) for new matrix", shape, data.Len())
 	}
 	return &Matrix{
 		shape: shape,
@@ -369,20 +333,20 @@ func sameShape(x, y []int) bool {
 
 // reshape implements binary rho
 // A⍴B: Array of shape A with data B
-func reshape(A, B Vector) Value {
-	if len(B) == 0 {
+func reshape(A, B *Vector) Value {
+	if B.Len() == 0 {
 		// Peculiar APL definition of reshape of empty vector: Use fill values.
 		B = NewIntVector(0)
 	}
-	if len(A) == 0 {
-		return Vector{}
+	if A.Len() == 0 {
+		return NewVector()
 	}
-	nelems := one
-	shape := make([]int, len(A))
-	for i := range A {
-		n, ok := A[i].Inner().(Int)
+	nelems := Int(1)
+	shape := make([]int, A.Len())
+	for i := range A.All() {
+		n, ok := A.At(i).Inner().(Int)
 		if !ok || n < 0 || maxInt < n {
-			Errorf("bad shape for rho: %s is not a small integer", A[i])
+			Errorf("bad shape for rho: %s is not a small integer", A.At(i))
 		}
 		nelems *= n
 		if nelems > maxInt {
@@ -390,16 +354,20 @@ func reshape(A, B Vector) Value {
 		}
 		shape[i] = int(n)
 	}
-	values := make([]Value, nelems)
-	n := copy(values, B)
-	// replicate as needed by doubling in values.
-	for n < len(values) {
-		n += copy(values[n:], values[:n])
+	t := newVectorEditor(int(nelems), nil)
+	blen := B.Len()
+	for i := 0; i < blen && i < int(nelems); i++ {
+		t.Set(i, B.At(i))
 	}
-	if len(A) == 1 {
-		return NewVector(values)
+	// replicate as needed
+	for i := blen; i < int(nelems); i++ {
+		t.Set(i, t.At(i-blen))
 	}
-	return NewMatrix(shape, NewVector(values))
+	v := t.Publish()
+	if A.Len() == 1 {
+		return v
+	}
+	return NewMatrix(shape, v)
 }
 
 // rotate returns a copy of v with elements rotated left by n.
@@ -408,19 +376,19 @@ func (m *Matrix) rotate(n int) Value {
 	if m.Rank() == 0 {
 		return &Matrix{}
 	}
-	elems := make([]Value, len(m.data))
+	elems := newVectorEditor(m.data.Len(), nil)
 	dim := m.shape[m.Rank()-1]
 	n %= dim
 	if n < 0 {
 		n += dim
 	}
-	pfor(true, dim, len(m.data)/dim, func(lo, hi int) {
+	pfor(true, dim, m.data.Len()/dim, func(lo, hi int) {
 		for i := lo; i < hi; i++ {
 			j := i * dim
-			doRotate(elems[j:j+dim], m.data[j:j+dim], n)
+			doRotate(elems, j, dim, m.data, j, n)
 		}
 	})
-	return NewMatrix(m.shape, elems)
+	return NewMatrix(m.shape, elems.Publish())
 }
 
 // vrotate returns a copy of v with elements rotated down by n.
@@ -433,37 +401,39 @@ func (m *Matrix) vrotate(n int) Value {
 		return m
 	}
 
-	elems := make([]Value, len(m.data))
-	dim := len(m.data) / m.shape[0]
+	elems := newVectorEditor(m.data.Len(), nil)
+	dim := m.data.Len() / m.shape[0]
 
 	n *= dim
-	n %= len(m.data)
+	n %= m.data.Len()
 	if n < 0 {
-		n += len(m.data)
+		n += m.data.Len()
 	}
 
-	pfor(true, dim, len(m.data)/dim, func(lo, hi int) {
+	pfor(true, dim, m.data.Len()/dim, func(lo, hi int) {
 		for i := lo; i < hi; i++ {
 			j := i * dim
-			n := (n + j) % len(m.data)
-			copy(elems[j:j+dim], m.data[n:n+dim])
+			n := (n + j) % m.data.Len()
+			for k := range dim {
+				elems.Set(j+k, m.data.At(n+k))
+			}
 		}
 	})
 
-	return NewMatrix(m.shape, elems)
+	return NewMatrix(m.shape, elems.Publish())
 }
 
 // transpose returns (as a new matrix) the transposition of the argument.
 func (m *Matrix) transpose(c Context) *Matrix {
 	// Fast version for common 2d case.
 	if len(m.shape) == 2 {
-		data := make([]Value, len(m.data))
+		data := newVectorEditor(m.data.Len(), nil)
 		xdim, ydim := m.shape[0], m.shape[1] // For new matrix.
-		pfor(true, 1, len(data), func(lo, hi int) {
+		pfor(true, 1, data.Len(), func(lo, hi int) {
 			nx := lo / ydim
 			ny := lo % ydim
-			for _, v := range m.data[lo:hi] {
-				data[ny*xdim+nx] = v
+			for _, v := range m.data.Slice(lo, hi) {
+				data.Set(ny*xdim+nx, v)
 				ny++
 				if ny >= ydim {
 					nx++
@@ -471,40 +441,37 @@ func (m *Matrix) transpose(c Context) *Matrix {
 				}
 			}
 		})
-		return NewMatrix([]int{ydim, xdim}, data)
+		return NewMatrix([]int{ydim, xdim}, data.Publish())
 	}
-	nShape := make(Vector, len(m.shape))
+	nShape := newVectorEditor(len(m.shape), nil)
 	origin := c.Config().Origin()
-	for i := range nShape {
-		nShape[len(nShape)-1-i] = Int(i + origin)
+	for i := range nShape.Len() {
+		nShape.Set(len(m.shape)-1-i, Int(i+origin))
 	}
-	return m.binaryTranspose(c, nShape)
+	return m.binaryTranspose(c, nShape.Publish())
 }
 
 // binaryTranspose returns the transposition of m specified by v,
 // defined by (v transp m)[i] = m[i[v]] (i is in general an index vector).
 // APL calls this operator the dyadic transpose.
-func (m *Matrix) binaryTranspose(c Context, v Vector) *Matrix {
+func (m *Matrix) binaryTranspose(c Context, v *Vector) *Matrix {
 	origin := c.Config().Origin()
-	if len(v) != m.Rank() {
-		Errorf("transp: vector length %d != matrix rank %d", len(v), m.Rank())
+	if v.Len() != m.Rank() {
+		Errorf("transp: vector length %d != matrix rank %d", v.Len(), m.Rank())
 	}
 
 	// Extract old-to-new index mapping and determine rank.
-	oldToNew := make([]int, len(v))
+	oldToNew := make([]int, v.Len())
 	rank := -1
-	for i := range v {
-		vi, ok := v[i].(Int)
-		if !ok {
-			Errorf("transp: non-int index %v", v[i])
-		}
-		if vi < Int(origin) || vi >= Int(origin+m.Rank()) {
+	for i := range v.All() {
+		vi := v.intAt(i, "transp index")
+		if vi < origin || vi >= origin+m.Rank() {
 			Errorf("transp: out-of-range index %v", vi)
 		}
-		vi -= Int(origin)
-		oldToNew[i] = int(vi)
-		if rank <= int(vi) {
-			rank = int(vi) + 1
+		vi -= origin
+		oldToNew[i] = vi
+		if rank <= vi {
+			rank = vi + 1
 		}
 	}
 
@@ -529,8 +496,8 @@ func (m *Matrix) binaryTranspose(c Context, v Vector) *Matrix {
 	}
 
 	old := m.data
-	data := make([]Value, sz)
-	pfor(true, 1, len(data), func(lo, hi int) {
+	data := newVectorEditor(sz, nil)
+	pfor(true, 1, data.Len(), func(lo, hi int) {
 		// Compute starting index
 		index := make([]int, rank)
 		i := lo
@@ -543,11 +510,11 @@ func (m *Matrix) binaryTranspose(c Context, v Vector) *Matrix {
 		for i := lo; i < hi; i++ {
 			// Compute old index for this new entry.
 			oi := 0
-			for j := range v {
+			for j := range v.All() {
 				oi = oi*m.shape[j] + index[oldToNew[j]]
 			}
 
-			data[i] = old[oi]
+			data.Set(i, old.At(oi))
 
 			// Increment index.
 			for j := rank - 1; j >= 0; j-- {
@@ -559,7 +526,7 @@ func (m *Matrix) binaryTranspose(c Context, v Vector) *Matrix {
 		}
 	})
 
-	return NewMatrix(shape, data)
+	return NewMatrix(shape, data.Publish())
 }
 
 // catenate returns the catenation x, y, along the last axis.
@@ -575,40 +542,43 @@ func (x *Matrix) catenate(y *Matrix) *Matrix {
 		Errorf("rank 0 matrix for ,")
 	}
 	var shape []int
-	var data Vector
+	data := newVectorEditor(0, nil)
 	var nrows int
 	setShape := func(m *Matrix, extra int) {
-		shape = make([]int, m.Rank())
-		copy(shape, m.shape)
+		shape = slices.Clone(m.shape)
 		shape[len(shape)-1] += extra
 		nrows = size(shape[:len(shape)-1])
 	}
 	copyElems := func(nLeft, advLeft, nRight, advRight int) {
 		di, li, ri := 0, 0, 0
 		for i := 0; i < nrows; i++ {
-			copy(data[di:di+nLeft], x.data[li:li+nLeft])
+			for k := range nLeft {
+				data.Set(di+k, x.data.At(li+k))
+			}
 			di += nLeft
 			li += advLeft
-			copy(data[di:di+nRight], y.data[ri:ri+nRight])
+			for k := range nRight {
+				data.Set(di+k, y.data.At(ri+k))
+			}
 			di += nRight
 			ri += advRight
 		}
 	}
 	switch {
 	default:
-		Errorf("catenate shape mismatch: %d, %d", NewIntVector(x.shape...), NewIntVector(y.shape...))
+		Errorf("catenate shape mismatch: %v, %v", NewIntVector(x.shape...), NewIntVector(y.shape...))
 
 	case x.Rank() == y.Rank() && sameShape(x.shape[:len(x.shape)-1], y.shape[:len(y.shape)-1]):
 		// list, list
 		setShape(x, y.shape[len(y.shape)-1])
-		data = make(Vector, len(x.data)+len(y.data))
-		xsize, ysize := x.ElemSize(), y.ElemSize()
+		data.Resize(x.data.Len() + y.data.Len())
+		xsize, ysize := x.shape[len(x.shape)-1], y.shape[len(y.shape)-1]
 		copyElems(xsize, xsize, ysize, ysize)
 
 	case x.Rank() == y.Rank()+1 && sameShape(x.shape[:len(x.shape)-1], y.shape):
 		// list, elem
 		setShape(x, 1)
-		data = make(Vector, len(x.data)+len(y.data))
+		data.Resize(x.data.Len() + y.data.Len())
 		xsize := x.shape[len(x.shape)-1]
 		ysize := y.Size() / nrows
 		copyElems(xsize, xsize, ysize, ysize)
@@ -616,7 +586,7 @@ func (x *Matrix) catenate(y *Matrix) *Matrix {
 	case x.Rank()+1 == y.Rank() && sameShape(x.shape, y.shape[:len(y.shape)-1]):
 		// elem, list
 		setShape(y, 1)
-		data = make(Vector, len(y.data)+len(x.data))
+		data.Resize(y.data.Len() + x.data.Len())
 		xsize := x.Size() / nrows
 		ysize := y.shape[len(y.shape)-1]
 		copyElems(xsize, xsize, ysize, ysize)
@@ -624,18 +594,18 @@ func (x *Matrix) catenate(y *Matrix) *Matrix {
 	case x.Rank() == 1 && x.shape[0] == 1 && y.Rank() > 1:
 		// scalar extension, list
 		setShape(y, 1)
-		data = make(Vector, len(y.data)+nrows)
+		data.Resize(y.data.Len() + nrows)
 		ysize := y.shape[len(y.shape)-1]
 		copyElems(1, 0, ysize, ysize)
 
 	case x.Rank() > 1 && y.Rank() == 1 && y.shape[0] == 1:
 		// list, scalar extension
 		setShape(x, 1)
-		data = make(Vector, len(x.data)+nrows)
+		data.Resize(x.data.Len() + nrows)
 		xsize := x.shape[len(x.shape)-1]
 		copyElems(xsize, xsize, 1, 0)
 	}
-	return NewMatrix(shape, data)
+	return NewMatrix(shape, data.Publish())
 }
 
 // catenateFirst returns the catenation x, y, along the first axis.
@@ -644,74 +614,70 @@ func (x *Matrix) catenateFirst(y *Matrix) *Matrix {
 		Errorf("rank 0 matrix for ,%%")
 	}
 	var shape []int
-	var data Vector
+	var data *vectorEditor
 	switch {
 	default:
-		Errorf("catenateFirst shape mismatch: %d, %d", NewIntVector(x.shape...), NewIntVector(y.shape...))
+		Errorf("catenateFirst shape mismatch: %v, %v", NewIntVector(x.shape...), NewIntVector(y.shape...))
 
 	case x.Rank() == y.Rank() && sameShape(x.shape[1:], y.shape[1:]):
 		// list, list
-		shape = make([]int, x.Rank())
-		copy(shape, x.shape)
+		shape = slices.Clone(x.shape)
 		shape[0] = x.shape[0] + y.shape[0]
 
 	case x.Rank() == y.Rank()+1 && sameShape(x.shape[1:], y.shape):
 		// list, elem
-		shape = make([]int, x.Rank())
-		copy(shape, x.shape)
+		shape = slices.Clone(x.shape)
 		shape[0]++
 
 	case x.Rank()+1 == y.Rank() && sameShape(x.shape, y.shape[1:]):
 		// elem, list
-		shape = make([]int, y.Rank())
-		copy(shape, y.shape)
+		shape = slices.Clone(y.shape)
 		shape[0]++
 
 	case x.Rank() == 1 && x.shape[0] == 1 && y.Rank() > 1:
 		// scalar extension, list
-		shape = make([]int, y.Rank())
-		copy(shape, y.shape)
+		shape = slices.Clone(y.shape)
 		shape[0]++
+		a := x.data.At(0)
+		data = newVectorEditor(0, nil)
 		elem := y.ElemSize()
-		a := x.data[0]
-		data = make(Vector, elem+len(y.data))
-		for i := 0; i < elem; i++ {
-			data[i] = a
+		for range elem {
+			data.Append(a)
+		}
+		for _, e := range y.data.All() {
+			data.Append(e)
 		}
-		copy(data[elem:], y.data)
 
 	case x.Rank() > 1 && y.Rank() == 1 && y.shape[0] == 1:
 		// list, scalar extension
-		shape = make([]int, x.Rank())
-		copy(shape, x.shape)
+		shape = slices.Clone(x.shape)
 		shape[0]++
 		elem := x.ElemSize()
-		b := y.data[0]
-		data = make(Vector, elem+len(x.data))
-		copy(data, x.data)
-		ext := data[len(x.data):]
-		for i := 0; i < elem; i++ {
-			ext[i] = b
+		b := y.data.At(0)
+		data = x.data.edit()
+		for range elem {
+			data.Append(b)
 		}
 	}
 	if data == nil {
-		data = make(Vector, len(x.data)+len(y.data))
-		copy(data, x.data)
-		copy(data[len(x.data):], y.data)
+		data = x.data.edit()
+		for _, e := range y.data.All() {
+			data.Append(e)
+		}
 	}
-	return NewMatrix(shape, data)
+	return NewMatrix(shape, data.Publish())
 }
 
 // sel returns the selection of m according to v.
 // The selection applies to the final axis.
-func (m *Matrix) sel(c Context, v Vector) *Matrix {
+func (m *Matrix) sel(c Context, v *Vector) *Matrix {
 	// All lhs values must be small integers.
 	if !v.AllInts() {
 		Errorf("sel: left operand must be small integers")
 	}
 
 	var count int64
-	for _, x := range v {
+	for _, x := range v.All() {
 		x := x.(Int)
 		if x < 0 {
 			count -= int64(x)
@@ -719,15 +685,14 @@ func (m *Matrix) sel(c Context, v Vector) *Matrix {
 			count += int64(x)
 		}
 	}
-	if len(v) != 1 && len(v) != m.shape[len(m.shape)-1] {
-		Errorf("sel: bad length %d for shape %s", len(v), NewIntVector(m.shape...))
+	if v.Len() != 1 && v.Len() != m.shape[len(m.shape)-1] {
+		Errorf("sel: bad length %d for shape %s", v.Len(), NewIntVector(m.shape...))
 	}
-	if len(v) == 1 {
+	if v.Len() == 1 {
 		count *= int64(m.shape[len(m.shape)-1])
 	}
 
-	shape := make([]int, len(m.shape))
-	copy(shape, m.shape)
+	shape := slices.Clone(m.shape)
 	shape[len(shape)-1] = int(count)
 
 	for _, dim := range shape[:len(shape)-1] {
@@ -737,46 +702,45 @@ func (m *Matrix) sel(c Context, v Vector) *Matrix {
 		Errorf("sel: result too large: %d elements", count)
 	}
 
-	result := make(Vector, 0, count)
-	for i, y := range m.data {
-		c := v[i%len(v)].(Int)
+	result := newVectorEditor(0, nil)
+	for i, y := range m.data.All() {
+		c := v.At(i % v.Len()).(Int)
 		if c < 0 {
 			c = -c
 			y = zero
 		}
 		for ; c > 0; c-- {
-			result = append(result, y)
+			result.Append(y)
 		}
 	}
 
-	return NewMatrix(shape, result)
+	return NewMatrix(shape, result.Publish())
 }
 
 // take returns v take m.
-func (m *Matrix) take(c Context, v Vector) *Matrix {
+func (m *Matrix) take(c Context, v *Vector) *Matrix {
 	if !v.AllInts() {
 		Errorf("take: left operand must be small integers")
 	}
 	// Extend short vector to full rank using shape.
-	if len(v) > m.Rank() {
+	if v.Len() > m.Rank() {
 		// Rank mismatch, but if m is of unit size, we can just raise its rank.
 		if m.Size() != 1 {
 			Errorf("take: bad rank %d for shape %s", m.Rank(), v)
 		}
 		// Create a 1x1x1... matrix and use that as the argument.
-		shape := make([]int, len(v))
+		shape := make([]int, v.Len())
 		for i := range shape {
 			shape[i] = 1
 		}
 		m = NewMatrix(shape, m.data)
 	}
-	if len(v) < m.Rank() {
-		ext := make(Vector, m.Rank())
-		copy(ext, v)
-		for i := len(v); i < m.Rank(); i++ {
-			ext[i] = Int(m.shape[i])
+	if v.Len() < m.Rank() {
+		ext := v.edit()
+		for ext.Len() < m.Rank() {
+			ext.Append(Int(m.shape[ext.Len()]))
 		}
-		v = ext
+		v = ext.Publish()
 	}
 
 	// Compute new shape.
@@ -789,7 +753,7 @@ func (m *Matrix) take(c Context, v Vector) *Matrix {
 	// origin is the location, in result space, of the upper left corner of the full m.
 	origin := make([]int, m.Rank())
 	count := int64(1) // Number of elements in result.
-	for i, x := range v {
+	for i, x := range v.All() {
 		var mb pos
 		var o int
 		y := int(x.(Int))
@@ -815,8 +779,8 @@ func (m *Matrix) take(c Context, v Vector) *Matrix {
 	// TODO Is there a faster way?
 	fill := fillValue(m.data)
 	rCoords := make([]int, len(shape)) // Matrix coordinates in result.
-	result := make(Vector, count, count)
-	for i := range result {
+	result := newVectorEditor(int(count), nil)
+	for i := range result.Len() {
 		inside := true
 		mi := 0
 		// See if this location is inside the bounding box for m.
@@ -831,9 +795,9 @@ func (m *Matrix) take(c Context, v Vector) *Matrix {
 			mi += loc
 		}
 		if inside {
-			result[i] = m.data[mi] // TODO
+			result.Set(i, m.data.At(mi)) // TODO
 		} else {
-			result[i] = fill
+			result.Set(i, fill)
 		}
 		// Increment destination indexes.
 		for k := len(rCoords) - 1; k >= 0; k-- {
@@ -844,45 +808,72 @@ func (m *Matrix) take(c Context, v Vector) *Matrix {
 			rCoords[k] = 0
 		}
 	}
-	return NewMatrix(shape, result)
+	return NewMatrix(shape, result.Publish())
+}
+
+// partition returns a vector of the subblocks of m, selected and grouped
+// by the values in score. Subblocks with score 0 are ignored.
+// Subblocks with non-zero score are included, grouped with boundaries
+// at every point where the score exceeds the previous score.
+func (m *Matrix) partition(scoreM *Matrix) Value {
+	if len(scoreM.shape) != 1 {
+		Errorf("part: left argument must be scalar or vector")
+	}
+	score := scoreM.data
+	lastDim := m.shape[len(m.shape)-1]
+	if scoreM.shape[0] == 1 {
+		// Make a new score the width of the matrix.
+		n := score.uintAt(0, "part: score")
+		x := make([]Value, lastDim)
+		for i := range x {
+			x[i] = Int(n)
+		}
+		score = NewVector(x...)
+	}
+	if score.Len() != lastDim {
+		Errorf("part: length mismatch")
+	}
+	res, dim := m.data.doPartition(score)
+	newShape := append([]int{}, m.shape...)
+	newShape[len(newShape)-1] = dim
+	return NewMatrix(newShape, res)
 }
 
 // drop returns v drop m.
-func (m *Matrix) drop(c Context, v Vector) *Matrix {
+func (m *Matrix) drop(c Context, v *Vector) *Matrix {
 	// Extend short vector to full rank using zeros.
-	if len(v) > m.Rank() {
+	if v.Len() > m.Rank() {
 		Errorf("take: argument %v too large for matrix with shape %s", v, NewIntVector(m.shape...))
 	}
 	if !v.AllInts() {
 		Errorf("drop: left operand must be small integers")
 	}
-	if len(v) < m.Rank() {
-		ext := make(Vector, m.Rank())
-		copy(ext, v)
-		for i := len(v); i < m.Rank(); i++ {
-			ext[i] = zero
+	if v.Len() < m.Rank() {
+		ext := v.edit()
+		for range m.Rank() - v.Len() {
+			ext.Append(zero)
 		}
-		v = ext
+		v = ext.Publish()
 	}
 
 	// All lhs values must be small integers in range for m's shape.
 	// Convert to parameters for take.
 	//	1 drop x = (1 - N) take x
 	//	-1 drop x = (N - 1) take x
-	take := make(Vector, len(v))
-	for i, x := range v {
+	take := v.edit()
+	for i, x := range v.All() {
 		x := int(x.(Int))
 		switch {
 		case x < -m.shape[i], x > m.shape[i]:
-			take[i] = zero
+			take.Set(i, zero)
 		case x >= 0:
-			take[i] = Int(x - m.shape[i])
+			take.Set(i, Int(x-m.shape[i]))
 		case x < 0:
-			take[i] = Int(m.shape[i] + x)
+			take.Set(i, Int(m.shape[i]+x))
 		}
 	}
 
-	return m.take(c, take)
+	return m.take(c, take.Publish())
 }
 
 // split reduces the matrix by one dimension.
@@ -892,30 +883,30 @@ func (m *Matrix) split() Value {
 		Errorf("cannot split rank %d matrix", len(m.shape))
 	}
 	// Matrix of vectors.
-	n := m.shape[len(m.shape)-1]
-	mData := make([]Value, 0, size(m.shape[:len(m.shape)-1]))
-	for i := 0; i < len(m.data); i += n {
-		mData = append(mData, NewVector(m.data[i:i+n]))
+	shape, n := m.shape[:len(m.shape)-1], m.shape[len(m.shape)-1]
+	mData := newVectorEditor(size(shape), nil)
+	for i := range mData.Len() {
+		mData.Set(i, NewVectorSeq(m.data.Slice(i*n, (i+1)*n)))
 	}
-	return NewMatrix(m.shape[:len(m.shape)-1], mData).shrink()
+	return NewMatrix(shape, mData.Publish()).shrink()
 }
 
 // mix builds a matrix from the elements of the nested matrix.
 func (m *Matrix) mix(c Context) Value {
 	// If it's all scalar, nothing to do.
-	if allScalars(m.data) {
-		return m.Copy()
+	if m.data.allScalars() {
+		return m
 	}
 	shape := []int{0}
-	for _, e := range m.data {
+	for _, e := range m.data.All() {
 		switch e := e.(type) {
 		default:
 			if shape[len(shape)-1] == 0 {
 				shape[len(shape)-1] = 1
 			}
-		case Vector:
-			if shape[len(shape)-1] < len(e) {
-				shape[len(shape)-1] = len(e)
+		case *Vector:
+			if shape[len(shape)-1] < e.Len() {
+				shape[len(shape)-1] = e.Len()
 			}
 		case *Matrix:
 			for len(e.shape) > len(shape) {
@@ -928,9 +919,9 @@ func (m *Matrix) mix(c Context) Value {
 			}
 		}
 	}
-	var data []Value
+	data := newVectorEditor(0, nil)
 	vshape := NewIntVector(shape...)
-	for _, e := range m.data {
+	for _, e := range m.data.All() {
 		var nm *Matrix
 		takeShape := make([]int, len(shape))
 		for i := range takeShape {
@@ -938,12 +929,12 @@ func (m *Matrix) mix(c Context) Value {
 		}
 		switch e := e.(type) {
 		default:
-			nm = NewMatrix(takeShape, []Value{e}).take(c, vshape)
-		case Vector:
-			takeShape[len(takeShape)-1] = len(e)
+			nm = NewMatrix(takeShape, NewVector(e)).take(c, vshape)
+		case *Vector:
+			takeShape[len(takeShape)-1] = e.Len()
 			nm = NewMatrix(takeShape, e).take(c, vshape)
 		case *Matrix:
-			offset := len(vshape) - len(e.shape)
+			offset := vshape.Len() - len(e.shape)
 			same := offset == 0
 			for i := range e.shape {
 				if e.shape[i] != takeShape[offset+i] {
@@ -956,28 +947,31 @@ func (m *Matrix) mix(c Context) Value {
 				nm = NewMatrix(takeShape, e.data).take(c, vshape)
 			}
 		}
-		data = append(data, nm.data...)
+		for _, elem := range nm.data.All() {
+			data.Append(elem)
+		}
 	}
-	return NewMatrix(append(m.shape, shape...), data)
+	return NewMatrix(append(m.shape, shape...), data.Publish())
 }
 
 // grade returns as a Vector the indexes that sort the rows of m
 // into increasing order.
-func (m *Matrix) grade(c Context) Vector {
+func (m *Matrix) grade(c Context) *Vector {
 	x := make([]int, m.shape[0])
 	for i := range x {
 		x[i] = i
 	}
 	v := m.data
-	stride := len(v) / m.shape[0]
+	stride := v.Len() / m.shape[0]
 	sort.Slice(x, func(i, j int) bool {
 		i = x[i] * stride
 		j = x[j] * stride
 		for k := 0; k < stride; k++ {
-			if toBool(c.EvalBinary(v[i+k], "==", v[j+k])) {
+			cmp := OrderedCompare(c, v.At(i+k), v.At(j+k))
+			if cmp == 0 {
 				continue
 			}
-			return toBool(c.EvalBinary(v[i+k], "<", v[j+k]))
+			return cmp < 0
 		}
 		return false
 	})
@@ -991,15 +985,17 @@ func (m *Matrix) grade(c Context) Vector {
 // inverse returns the matrix inverse of m. Note: although the code forbids
 // non-scalar elements, they actually "work", but they are probably more confusing
 // than helpful:
-//   x = 2 2 rho 1 2 3 4; x[1;1]=2 3; inv x
-//      (2 2/3)   (-1 -1/3)
-//    (-3/2 -1/2)     (1 1/2)
-//  x+.*inv x
-//    (1 1) (0 0)
-//    (0 0) (1 1)
-//  inv inv x # This one is clearly nuts.
-//    (2 3) (2 2)
-//    (3 3) (4 4)
+//
+//	x = 2 2 rho 1 2 3 4; x[1;1]=2 3; inv x
+//	      (2 2/3)   (-1 -1/3)
+//	  (-3/2 -1/2)     (1 1/2)
+//	x+.*inv x
+//	  (1 1) (0 0)
+//	  (0 0) (1 1)
+//	inv inv x # This one is clearly nuts.
+//	  (2 3) (2 2)
+//	  (3 3) (4 4)
+//
 // So they are forbidden.
 func (m *Matrix) inverse(c Context) Value {
 	const (
@@ -1010,9 +1006,11 @@ func (m *Matrix) inverse(c Context) Value {
 	case 0:
 		Errorf("inverse of empty matrix")
 	case 1:
-		return NewMatrix(m.shape, NewVector(m.data).inverse(c).(Vector))
+		return NewMatrix(m.shape, m.data.inverse(c).(*Vector))
 	case 2:
 		// OK
+	case 3:
+		Errorf("cannot compute inverse of matrix with dimension > 2")
 	}
 	dim := m.shape[0]
 	if m.shape[1] != dim {
@@ -1029,7 +1027,7 @@ func (m *Matrix) inverse(c Context) Value {
 	for y := 0; y < dim; y++ {
 		row := t[y]
 		for x := 0; x < dim; x++ {
-			row[x] = m.data[i]
+			row[x] = m.data.At(i)
 			i++
 			if x%dim == y {
 				row[dim+x] = one
@@ -1091,9 +1089,9 @@ func (m *Matrix) inverse(c Context) Value {
 		}
 	}
 	// Now extract the right hand side of the working area.
-	data := make([]Value, 0, len(m.data))
+	data := newVectorEditor(0, nil)
 	for _, row := range t {
-		data = append(data, row[dim:]...)
+		data.Append(row[dim:]...)
 	}
-	return NewMatrix(m.shape, data)
+	return NewMatrix(m.shape, data.Publish())
 }
