@@ -5,8 +5,9 @@
 package value
 
 import (
+	"crypto/rand"
+	"fmt"
 	"math/big"
-	"unicode/utf8"
 )
 
 // Unary operators.
@@ -52,8 +53,17 @@ func bigFloatWrap(op func(*big.Float, *big.Float) *big.Float) func(Context, *big
 
 // bigIntRand sets a to a random number in [origin, origin+b].
 func bigIntRand(c Context, a, b *big.Int) *big.Int {
-	a.Rand(c.Config().Random(), b)
-	return a.Add(a, c.Config().BigOrigin())
+	config := c.Config()
+	config.LockRandom()
+	for {
+		i, err := rand.Int(config.Source(), b)
+		if err == nil {
+			a.Set(i)
+			break
+		}
+	}
+	config.UnlockRandom()
+	return a
 }
 
 func self(c Context, v Value) Value {
@@ -75,7 +85,7 @@ func realPhase(c Context, v Value) Value {
 // v must be a scalar.
 func vectorSelf(c Context, v Value) Value {
 	switch v.(type) {
-	case Vector:
+	case *Vector:
 		Errorf("internal error: vectorSelf of vector")
 	case *Matrix:
 		Errorf("internal error: vectorSelf of matrix")
@@ -108,12 +118,16 @@ func floatSelf(c Context, v Value) BigFloat {
 // text returns a vector of Chars holding the string representation
 // of the value.
 func text(c Context, v Value) Value {
-	str := v.Sprint(c.Config())
-	elem := make([]Value, utf8.RuneCountInString(str))
-	for i, r := range str {
-		elem[i] = Char(r)
+	return newCharVector(v.Sprint(c.Config()))
+}
+
+// newCharVector takes a string and returns its representation as a Vector of Chars.
+func newCharVector(s string) Value {
+	edit := newVectorEditor(0, nil)
+	for _, r := range s {
+		edit.Append(Char(r))
 	}
-	return NewVector(elem)
+	return edit.Publish()
 }
 
 // Implemented in package run, handled as a func to avoid a dependency loop.
@@ -121,6 +135,34 @@ var IvyEval func(context Context, s string) Value
 
 var UnaryOps = make(map[string]UnaryOp)
 
+func printValue(c Context, v Value) Value {
+	fmt.Fprintln(c.Config().Output(), v.Sprint(c.Config()))
+	return QuietValue{v}
+}
+
+// bigFloatRand returns a uniformly distributed BigFloat in the range [0, f).
+// For [0, 1), the mean should be 0.5 and 𝛔 should be 1/√12, or 0.2887.
+// A test of a million values yielded 0.500161824174 0.288777488704.
+func bigFloatRand(c Context, f *big.Float) Value {
+	if f.Sign() < 0 {
+		Errorf("rand of negative number")
+	}
+	// We want two integers FloatPrec bits long.
+	config := c.Config()
+	max := big.NewInt(1)
+	max.Lsh(max, config.FloatPrec())
+	max.Add(max, bigIntOne.Int) // We want range [0,n).
+	// Now grab a random integer from [0, max)
+	randInt := big.NewInt(0)
+	bigIntRand(c, randInt, max)
+	// Make it a float and normalize it.
+	x := big.NewFloat(0).SetInt(randInt)
+	x.Quo(x, big.NewFloat(0).SetInt(max))
+	// Finally, scale it up to [0, f).
+	x.Mul(x, f)
+	return BigFloat{x}
+}
+
 func init() {
 	ops := []*unaryOp{
 		{
@@ -132,7 +174,10 @@ func init() {
 					if i <= 0 {
 						Errorf("illegal roll value %v", v)
 					}
-					return Int(c.Config().Origin()) + Int(c.Config().Random().Int63n(i))
+					c.Config().LockRandom()
+					res := Int(c.Config().Random().Int64N(int64(i)))
+					c.Config().UnlockRandom()
+					return Int(c.Config().Origin()) + res
 				},
 				bigIntType: func(c Context, v Value) Value {
 					if v.(BigInt).Sign() <= 0 {
@@ -143,26 +188,51 @@ func init() {
 			},
 		},
 
+		{
+			name:        "rand",
+			elementwise: true,
+			fn: [numType]unaryFn{
+				intType: func(c Context, v Value) Value {
+					var x big.Float
+					x.SetInt64(int64(v.(Int)))
+					return bigFloatRand(c, &x)
+				},
+				bigIntType: func(c Context, v Value) Value {
+					var x big.Float
+					x.SetInt(v.(BigInt).Int)
+					return bigFloatRand(c, &x)
+				},
+				bigRatType: func(c Context, v Value) Value {
+					var x big.Float
+					x.SetRat(v.(BigRat).Rat)
+					return bigFloatRand(c, &x)
+				},
+				bigFloatType: func(c Context, v Value) Value {
+					return bigFloatRand(c, v.(BigFloat).Float)
+				},
+			},
+		},
+
 		{
 			name:        "j",
 			elementwise: true,
 			fn: [numType]unaryFn{
 				intType: func(c Context, v Value) Value {
-					return NewComplex(zero, v)
+					return NewComplex(zero, v).shrink()
 				},
 				bigIntType: func(c Context, v Value) Value {
-					return NewComplex(zero, v)
+					return NewComplex(zero, v).shrink()
 				},
 				bigRatType: func(c Context, v Value) Value {
-					return NewComplex(zero, v)
+					return NewComplex(zero, v).shrink()
 				},
 				bigFloatType: func(c Context, v Value) Value {
-					return NewComplex(zero, v)
+					return NewComplex(zero, v).shrink()
 				},
 				complexType: func(c Context, v Value) Value {
 					// Multiply by i.
 					u := v.(Complex)
-					return NewComplex(c.EvalUnary("-", u.imag), u.real)
+					return NewComplex(c.EvalUnary("-", u.imag), u.real).shrink()
 				},
 			},
 		},
@@ -191,8 +261,8 @@ func init() {
 				bigFloatType: self,
 				complexType:  self,
 				vectorType: func(c Context, v Value) Value {
-					vv := v.(Vector)
-					return NewMatrix([]int{len(vv)}, vv).mix(c).shrink()
+					vv := v.(*Vector)
+					return NewMatrix([]int{vv.Len()}, vv).mix(c).shrink()
 
 				},
 				matrixType: func(c Context, v Value) Value {
@@ -231,7 +301,7 @@ func init() {
 					return unaryBigFloatOp(c, bigFloatWrap((*big.Float).Neg), v)
 				},
 				complexType: func(c Context, v Value) Value {
-					return v.(Complex).neg(c)
+					return v.(Complex).neg(c).shrink()
 				},
 			},
 		},
@@ -253,7 +323,7 @@ func init() {
 					return v.(BigFloat).inverse()
 				},
 				complexType: func(c Context, v Value) Value {
-					return v.(Complex).inverse(c)
+					return v.(Complex).inverse(c).shrink()
 				},
 			},
 		},
@@ -275,10 +345,10 @@ func init() {
 					return v.(BigFloat).inverse()
 				},
 				complexType: func(c Context, v Value) Value {
-					return v.(Complex).inverse(c)
+					return v.(Complex).inverse(c).shrink()
 				},
 				vectorType: func(c Context, v Value) Value {
-					return v.(Vector).inverse(c)
+					return v.(*Vector).inverse(c)
 				},
 				matrixType: func(c Context, v Value) Value {
 					return v.(*Matrix).inverse(c)
@@ -324,7 +394,7 @@ func init() {
 				bigFloatType: self,
 				complexType: func(c Context, v Value) Value {
 					u := v.(Complex)
-					return NewComplex(u.real, c.EvalUnary("-", u.imag)).shrink()
+					return NewComplex(u.real, c.EvalUnary("-", u.imag)).shrink().shrink()
 				},
 			},
 		},
@@ -411,7 +481,7 @@ func init() {
 					return unaryBigFloatOp(c, bigFloatWrap((*big.Float).Abs), v)
 				},
 				complexType: func(c Context, v Value) Value {
-					return v.(Complex).abs(c)
+					return v.(Complex).abs(c).shrink()
 				},
 			},
 		},
@@ -453,7 +523,7 @@ func init() {
 				bigRatType:   realPhase,
 				bigFloatType: realPhase,
 				complexType: func(c Context, v Value) Value {
-					return v.(Complex).phase(c)
+					return v.(Complex).phase(c).shrink()
 				},
 			},
 		},
@@ -555,27 +625,20 @@ func init() {
 			name: "iota",
 			fn: [numType]unaryFn{
 				intType: func(c Context, v Value) Value {
-					i := v.(Int)
-					if i < 0 || maxInt < i {
-						Errorf("bad iota %d", i)
-					}
-					if i == 0 {
-						return Vector{}
-					}
-					data := constIota(c.Config().Origin(), int(i))
-					n := make([]Value, i)
-					copy(n, data)
-					return NewVector(n)
+					return newIota(c.Config().Origin(), int(v.(Int)))
 				},
 				vectorType: func(c Context, v Value) Value {
 					// Produce a matrix of coordinates.
-					vv := v.(Vector)
-					if len(vv) == 0 {
+					vv := v.(*Vector)
+					if vv.Len() == 0 {
 						return empty
 					}
+					if vv.Len() == 1 {
+						return newIota(c.Config().Origin(), vv.intAt(0, "iota argument"))
+					}
 					nElems := 1
-					shape := make([]int, len(vv))
-					for i, coord := range vv {
+					shape := make([]int, vv.Len())
+					for i, coord := range vv.All() {
 						c, ok := coord.(Int)
 						if !ok || c < 0 || maxInt < c {
 							Errorf("bad coordinate in iota %s", coord)
@@ -587,13 +650,13 @@ func init() {
 						}
 					}
 					origin := c.Config().Origin()
-					elems := make([]Value, nElems)
-					counter := make([]int, len(vv))
+					elems := newVectorEditor(nElems, nil)
+					counter := make([]int, vv.Len())
 					for i := range counter {
 						counter[i] = origin
 					}
-					for i := range elems {
-						elems[i] = NewIntVector(counter...)
+					for i := range nElems {
+						elems.Set(i, NewIntVector(counter...))
 						for axis := len(counter) - 1; axis >= 0; axis-- {
 							counter[axis]++
 							if counter[axis]-origin < shape[axis] {
@@ -602,7 +665,7 @@ func init() {
 							counter[axis] = origin
 						}
 					}
-					return NewMatrix(shape, elems)
+					return NewMatrix(shape, elems.Publish())
 				},
 			},
 		},
@@ -629,7 +692,7 @@ func init() {
 					return empty
 				},
 				vectorType: func(c Context, v Value) Value {
-					return NewIntVector(len(v.(Vector)))
+					return NewIntVector(v.(*Vector).Len())
 				},
 				matrixType: func(c Context, v Value) Value {
 					return NewIntVector(v.(*Matrix).shape...)
@@ -659,7 +722,7 @@ func init() {
 					return one
 				},
 				vectorType: func(c Context, v Value) Value {
-					return Int(len(v.(Vector)))
+					return Int(v.(*Vector).Len())
 				},
 				matrixType: func(c Context, v Value) Value {
 					m := v.(*Matrix)
@@ -671,6 +734,67 @@ func init() {
 			},
 		},
 
+		{
+			name: "where",
+			fn: [numType]unaryFn{
+				intType: func(c Context, v Value) Value {
+					return empty
+				},
+				charType: func(c Context, v Value) Value {
+					return empty
+				},
+				bigIntType: func(c Context, v Value) Value {
+					return empty
+				},
+				bigRatType: func(c Context, v Value) Value {
+					return empty
+				},
+				bigFloatType: func(c Context, v Value) Value {
+					return empty
+				},
+				complexType: func(c Context, v Value) Value {
+					return empty
+				},
+				vectorType: func(c Context, v Value) Value {
+					vec := v.(*Vector)
+					result := newVectorEditor(0, nil)
+					origin := c.Config().Origin()
+					for i := range vec.Len() {
+						e := vec.uintAt(i, "where argument")
+						for range e {
+							result.Append(Int(origin + i))
+						}
+					}
+					return result.Publish()
+				},
+				matrixType: func(c Context, v Value) Value {
+					m := v.(*Matrix)
+					shape, vec := m.shape, m.data
+					result := newVectorEditor(0, nil)
+					coords := make([]int, len(shape)) // Zero-indexed.
+					origin := c.Config().Origin()
+					// Loop over the data in the matrix while odometer-counting
+					// the coordinates.
+					for i := range vec.Len() {
+						e := vec.uintAt(i, "where argument")
+						for range e {
+							c := newVectorEditor(len(shape), nil)
+							for i, x := range coords {
+								c.Set(i, Int(x+origin))
+							}
+							result.Append(c.Publish())
+						}
+						for j := len(coords) - 1; j >= 0; j-- {
+							if coords[j]++; coords[j] < shape[j] {
+								break
+							}
+							coords[j] = 0
+						}
+					}
+					return result.Publish()
+				},
+			},
+		},
 		{
 			name: "flatten",
 			fn: [numType]unaryFn{
@@ -693,14 +817,27 @@ func init() {
 					return oneElemVector(v)
 				},
 				vectorType: func(c Context, v Value) Value {
-					return NewVector(flatten(v.(Vector)))
+					return NewVectorSeq(flatten(v.(*Vector)))
 				},
 				matrixType: func(c Context, v Value) Value {
-					return c.EvalUnary("flatten", NewVector(v.(*Matrix).data))
+					return c.EvalUnary("flatten", v.(*Matrix).data)
 				},
 			},
 		},
 
+		{
+			name: "print",
+			fn: [numType]unaryFn{
+				intType:      printValue,
+				charType:     printValue,
+				bigIntType:   printValue,
+				bigRatType:   printValue,
+				bigFloatType: printValue,
+				vectorType:   printValue,
+				matrixType:   printValue,
+			},
+		},
+
 		{
 			name: ",",
 			fn: [numType]unaryFn{
@@ -712,7 +849,7 @@ func init() {
 				complexType:  vectorSelf,
 				vectorType:   self,
 				matrixType: func(c Context, v Value) Value {
-					return v.(*Matrix).data.Copy()
+					return v.(*Matrix).data
 				},
 			},
 		},
@@ -726,7 +863,7 @@ func init() {
 				bigRatType:   self,
 				bigFloatType: self,
 				vectorType: func(c Context, v Value) Value {
-					return v.(Vector).grade(c)
+					return v.(*Vector).grade(c)
 				},
 				matrixType: func(c Context, v Value) Value {
 					return v.(*Matrix).grade(c)
@@ -743,7 +880,7 @@ func init() {
 				bigRatType:   self,
 				bigFloatType: self,
 				vectorType: func(c Context, v Value) Value {
-					return v.(Vector).grade(c).reverse()
+					return v.(*Vector).grade(c).reverse()
 				},
 				matrixType: func(c Context, v Value) Value {
 					return v.(*Matrix).grade(c).reverse()
@@ -761,10 +898,10 @@ func init() {
 				bigFloatType: self,
 				complexType:  self,
 				vectorType: func(c Context, v Value) Value {
-					return v.(Vector).reverse()
+					return v.(*Vector).reverse()
 				},
 				matrixType: func(c Context, v Value) Value {
-					m := v.(*Matrix).Copy().(*Matrix)
+					m := v.(*Matrix)
 					if m.Rank() == 0 {
 						return m
 					}
@@ -773,13 +910,15 @@ func init() {
 					}
 					size := int(m.Size())
 					ncols := m.shape[m.Rank()-1]
-					x := m.data
+					x := m.data.edit()
 					for index := 0; index <= size-ncols; index += ncols {
 						for i, j := 0, ncols-1; i < j; i, j = i+1, j-1 {
-							x[index+i], x[index+j] = x[index+j], x[index+i]
+							xi, xj := x.At(index+i), x.At(index+j)
+							x.Set(index+i, xj)
+							x.Set(index+j, xi)
 						}
 					}
-					return m
+					return &Matrix{shape: m.shape, data: x.Publish()}
 				},
 			},
 		},
@@ -794,10 +933,10 @@ func init() {
 				bigFloatType: self,
 				complexType:  self,
 				vectorType: func(c Context, v Value) Value {
-					return v.(Vector).reverse()
+					return v.(*Vector).reverse()
 				},
 				matrixType: func(c Context, v Value) Value {
-					m := v.(*Matrix).Copy().(*Matrix)
+					m := v.(*Matrix)
 					if m.Rank() == 0 {
 						return m
 					}
@@ -806,17 +945,19 @@ func init() {
 					}
 					elemSize := int(m.ElemSize())
 					size := int(m.Size())
-					x := m.data
+					x := m.data.edit()
 					lo := 0
 					hi := size - elemSize
 					for lo < hi {
 						for i := 0; i < elemSize; i++ {
-							x[lo+i], x[hi+i] = x[hi+i], x[lo+i]
+							xl, xh := x.At(lo+i), x.At(hi+i)
+							x.Set(lo+i, xh)
+							x.Set(hi+i, xl)
 						}
 						lo += elemSize
 						hi -= elemSize
 					}
-					return m
+					return &Matrix{shape: m.shape, data: x.Publish()}
 				},
 			},
 		},
@@ -830,9 +971,7 @@ func init() {
 				bigRatType:   self,
 				bigFloatType: self,
 				complexType:  self,
-				vectorType: func(c Context, v Value) Value {
-					return v.(Vector).Copy()
-				},
+				vectorType:   self,
 				matrixType: func(c Context, v Value) Value {
 					m := v.(*Matrix)
 					if m.Rank() == 1 {
@@ -1042,6 +1181,7 @@ func init() {
 		{
 			name: "text",
 			fn: [numType]unaryFn{
+				charType:     func(c Context, v Value) Value { return text(c, v) },
 				intType:      func(c Context, v Value) Value { return text(c, v) },
 				bigIntType:   func(c Context, v Value) Value { return text(c, v) },
 				bigRatType:   func(c Context, v Value) Value { return text(c, v) },
@@ -1060,12 +1200,11 @@ func init() {
 					return IvyEval(c, string(char))
 				},
 				vectorType: func(c Context, v Value) Value {
-					text := v.(Vector)
+					text := v.(*Vector)
 					if !text.AllChars() {
 						Errorf("ivy: value is not a vector of char")
 					}
-					str, _ := text.oneLineSprint(c.Config(), !withParens, !withSpaces)
-					return IvyEval(c, str)
+					return IvyEval(c, text.Sprint(c.Config()))
 				},
 			},
 		},
@@ -1125,18 +1264,18 @@ func init() {
 				bigFloatType: self,
 				complexType:  self,
 				vectorType: func(c Context, v Value) Value {
-					u := v.(Vector)
-					if len(u) == 0 {
+					u := v.(*Vector)
+					if u.Len() == 0 {
 						return zero // TODO: If we add prototypes, this is a place it matters.
 					}
-					return u[0]
+					return u.At(0)
 				},
 				matrixType: func(c Context, v Value) Value {
 					u := v.(*Matrix).data
-					if len(u) == 0 {
+					if u.Len() == 0 {
 						return zero // TODO: If we add prototypes, this is a place it matters.
 					}
-					return u[0]
+					return u.At(0)
 				},
 			},
 		},
